import filecmp
import os
import unittest
from pathlib import Path

from sidermit import exceptions
from sidermit.city import graph
from sidermit.publictransportsystem import TransportModeManager, \
    TransportNetwork, Route


class TransportNetworkTest(unittest.TestCase):

    def setUp(self):
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        self.data_path = os.path.join(self.dir_path, 'file')
        self.data_path = os.path.join(self.data_path, 'transport_network')

    def test_routes(self):
        """
        test Route class
        :return:
        """

        mode_manager = TransportModeManager()
        bus_obj = mode_manager.get_mode("bus")

        r = Route("r1", bus_obj, "1,2,0,4,3", "3,4,0,2,1", "1,0,3", "3,0,1")

        self.assertTrue(isinstance(r, Route))

    def test_raises_routes_exceptions(self):
        """
        to check route class exceptions
        :return:
        """

        mode_manager = TransportModeManager()
        bus_obj = mode_manager.get_mode("bus")

        with self.assertRaises(exceptions.RouteIdIsNotValidException):
            Route(None, bus_obj, "1,2,0,4,3", "3,4,0,2,1", "1,0,3",
                  "3,0,1")

        with self.assertRaises(exceptions.ModeIsNotValidException):
            Route("r1", "train", "1,2,0,4,3", "3,4,0,2,1",
                  "1,0,3", "3,0,1")

        with self.assertRaises(exceptions.StopsSequencesException):
            Route("r1", bus_obj, "1,2,0,4,3", "3,4,0,2,1", "1,5,3",
                  "3,0,1")

        with self.assertRaises(exceptions.FirstStopIsNotValidException):
            Route("r1", bus_obj, "1,2,0,4,3", "3,4,0,2,1", "2,0,3",
                  "3,0,1")

        with self.assertRaises(exceptions.LastStopIsNotValidException):
            Route("r1", bus_obj, "1,2,0,4,3", "3,4,0,2,1", "1,0,2",
                  "3,0,1")

        with self.assertRaises(exceptions.NotCycleException):
            Route("r1", bus_obj, "1,2,0,4,3", "4,0,2,1", "1,0,3",
                  "4,0,1")

    def test_add_route_transport_network(self):
        """
        to test add_route method of transport_network class
        :return:
        """
        graph_obj = graph.Graph.build_from_parameters(5, 1000, 0.5, 2)
        mode_manager = TransportModeManager()
        bus_obj = mode_manager.get_mode("bus")
        network = TransportNetwork(graph_obj)
        route = Route("r1", bus_obj, "1,2,0,4,3", "3,4,0,2,1", "1,0,3", "3,0,1")
        network.add_route(route)

        self.assertTrue(isinstance(network.get_route("r1"), Route))
        self.assertEqual(len(network.get_routes()), 1)

        with self.assertRaises(exceptions.RouteIsNotvalidException):
            network.add_route("route")

        with self.assertRaises(exceptions.NodeSequencesIsNotValidException):
            route = Route("r2", bus_obj, "1,3,0,4,3", "3,4,0,2,1", "1,0,3",
                          "3,0,1")
            network.add_route(route)

        with self.assertRaises(exceptions.RouteIdDuplicatedException):
            route = Route("r1", bus_obj, "1,2,0,4,3", "3,4,0,2,1", "1,0,3",
                          "3,0,1")
            network.add_route(route)

    def test_remove_route_transport_network(self):
        """
        to test delete_route method of class transport_network class
        :return:
        """

        graph_obj = graph.Graph.build_from_parameters(5, 1000, 0.5, 2)
        mode_manager = TransportModeManager()
        bus_obj = mode_manager.get_mode("bus")
        network = TransportNetwork(graph_obj)
        route = Route("r1", bus_obj, "1,2,0,4,3", "3,4,0,2,1", "1,0,3", "3,0,1")
        network.add_route(route)
        network.remove_route("r1")
        self.assertEqual(len(network.get_routes()), 0)

        with self.assertRaises(exceptions.RouteIdNotFoundException):
            network.remove_route("r1")

    def test_network_to_file(self):
        """
        to test method to write a file with network information
        :return:
        """

        graph_obj = graph.Graph.build_from_parameters(5, 1000, 0.5, 2)
        mode_manager = TransportModeManager()
        bus_obj = mode_manager.get_mode("bus")
        metro = mode_manager.get_mode("metro")
        network = TransportNetwork(graph_obj)
        route1 = Route("r1", bus_obj, "1,2,0,4,3", "3,4,0,2,1", "1,0,3", "3,0,1")
        route2 = Route("r2", metro, "1,2,0,4,3", "3,4,0,2,1", "1,0,3", "3,0,1")
        network.add_route(route1)
        network.add_route(route2)
        # write file
        network.routes_to_file(os.path.join(self.data_path, 'write_test.csv'))
        file_obj = Path(os.path.join(self.data_path, 'write_test.csv'))
        # test
        self.assertTrue(file_obj.is_file())
        # to compare file with a test file
        self.assertTrue(
            filecmp.cmp(os.path.join(self.data_path, 'write1_test.csv'),
                        os.path.join(self.data_path, "write_test.csv")))
        # remove file
        os.remove(os.path.join(self.data_path, 'write_test.csv'))

    def test_raises_get_route_exceptions(self):
        """
        raises get_route exceptions from transport network class
        :return:
        """
        with self.assertRaises(exceptions.RouteIdNotFoundException):
            graph_obj = graph.Graph.build_from_parameters(5, 1000, 0.5, 2)
            network = TransportNetwork(graph_obj)
            network.get_route("r1")

    def test_get_feeder_routes(self):
        """
        to test add feeder routes
        :return:
        """
        graph_obj = graph.Graph.build_from_parameters(5, 1000, 0.5, 2)
        mode_manager = TransportModeManager()
        bus_obj = mode_manager.get_mode("bus")
        network = TransportNetwork(graph_obj)

        routes = network.get_feeder_routes(bus_obj)

        self.assertEqual(len(routes), 5)
        self.assertEqual(routes[0].nodes_sequence_i, [1, 2])
        self.assertEqual(routes[0].nodes_sequence_r, [2, 1])
        self.assertEqual(routes[0].stops_sequence_i, [1, 2])
        self.assertEqual(routes[0].stops_sequence_r, [2, 1])
        self.assertEqual(routes[1].nodes_sequence_i, [3, 4])
        self.assertEqual(routes[1].nodes_sequence_r, [4, 3])
        self.assertEqual(routes[1].stops_sequence_i, [3, 4])
        self.assertEqual(routes[1].stops_sequence_r, [4, 3])

        with self.assertRaises(exceptions.ModeIsNotValidException):
            network.get_feeder_routes("bus_obj")

    def test_get_circular_routes(self):
        """
        to test add circular routes
        :return:
        """
        g = graph.Graph.build_from_parameters(5, 1000, 0.5, 2)
        mode_manager = TransportModeManager()
        bus_obj = mode_manager.get_mode("bus")
        network = TransportNetwork(g)

        routes = network.get_circular_routes(bus_obj)

        self.assertEqual(len(routes), 2)
        self.assertEqual(routes[0].nodes_sequence_i, [2, 4, 6, 8, 10, 2])
        self.assertEqual(routes[0].stops_sequence_i, [2, 4, 6, 8, 10, 2])
        self.assertEqual(routes[0].nodes_sequence_r, [])
        self.assertEqual(routes[0].stops_sequence_r, [])

        self.assertEqual(routes[1].nodes_sequence_i, [])
        self.assertEqual(routes[1].stops_sequence_i, [])
        self.assertEqual(routes[1].nodes_sequence_r, [10, 8, 6, 4, 2, 10])
        self.assertEqual(routes[1].stops_sequence_r, [10, 8, 6, 4, 2, 10])

        with self.assertRaises(exceptions.ModeIsNotValidException):
            g = graph.Graph.build_from_parameters(5, 1000, 0.5, 2)
            network = TransportNetwork(g)
            network.get_circular_routes("bus_obj")

        with self.assertRaises(exceptions.CircularRouteIsNotValidException):
            g = graph.Graph.build_from_parameters(1, 1000, 0.5, 2)
            mode_manager = TransportModeManager()
            bus_obj = mode_manager.get_mode("bus")
            network = TransportNetwork(g)
            network.get_circular_routes(bus_obj)

    def test_get_radial_routes(self):
        """
        to test add radial routes
        :return:
        """
        g = graph.Graph.build_from_parameters(5, 1000, 0.5, 2)
        mode_manager = TransportModeManager()
        bus_obj = mode_manager.get_mode("bus")
        metro_obj = mode_manager.get_mode("metro")
        network = TransportNetwork(g)

        routes = network.get_radial_routes(mode_obj=bus_obj, short=False, express=False)

        self.assertEqual(len(routes), 5)
        self.assertEqual(routes[0].nodes_sequence_i, [1, 2, 0])
        self.assertEqual(routes[0].nodes_sequence_r, [0, 2, 1])
        self.assertEqual(routes[0].stops_sequence_i, [1, 2, 0])
        self.assertEqual(routes[0].stops_sequence_r, [0, 2, 1])
        self.assertEqual(routes[1].nodes_sequence_i, [3, 4, 0])
        self.assertEqual(routes[1].nodes_sequence_r, [0, 4, 3])
        self.assertEqual(routes[1].stops_sequence_i, [3, 4, 0])
        self.assertEqual(routes[1].stops_sequence_r, [0, 4, 3])

        routes = network.get_radial_routes(mode_obj=bus_obj, short=True, express=False)

        self.assertEqual(len(routes), 5)
        self.assertEqual(routes[0].nodes_sequence_i, [2, 0])
        self.assertEqual(routes[0].nodes_sequence_r, [0, 2])
        self.assertEqual(routes[0].stops_sequence_i, [2, 0])
        self.assertEqual(routes[0].stops_sequence_r, [0, 2])
        self.assertEqual(routes[1].nodes_sequence_i, [4, 0])
        self.assertEqual(routes[1].nodes_sequence_r, [0, 4])
        self.assertEqual(routes[1].stops_sequence_i, [4, 0])
        self.assertEqual(routes[1].stops_sequence_r, [0, 4])

        routes = network.get_radial_routes(mode_obj=bus_obj, short=False, express=True)

        self.assertEqual(len(routes), 5)
        self.assertEqual(routes[0].nodes_sequence_i, [1, 2, 0])
        self.assertEqual(routes[0].nodes_sequence_r, [0, 2, 1])
        self.assertEqual(routes[0].stops_sequence_i, [1, 0])
        self.assertEqual(routes[0].stops_sequence_r, [0, 1])
        self.assertEqual(routes[1].nodes_sequence_i, [3, 4, 0])
        self.assertEqual(routes[1].nodes_sequence_r, [0, 4, 3])
        self.assertEqual(routes[1].stops_sequence_i, [3, 0])
        self.assertEqual(routes[1].stops_sequence_r, [0, 3])

        routes = network.get_radial_routes(mode_obj=metro_obj, short=True, express=True)

        self.assertEqual(len(routes), 5)
        self.assertEqual(routes[0].nodes_sequence_i, [2, 0])
        self.assertEqual(routes[0].nodes_sequence_r, [0, 2])
        self.assertEqual(routes[0].stops_sequence_i, [2, 0])
        self.assertEqual(routes[0].stops_sequence_r, [0, 2])
        self.assertEqual(routes[1].nodes_sequence_i, [4, 0])
        self.assertEqual(routes[1].nodes_sequence_r, [0, 4])
        self.assertEqual(routes[1].stops_sequence_i, [4, 0])
        self.assertEqual(routes[1].stops_sequence_r, [0, 4])

        with self.assertRaises(exceptions.ModeIsNotValidException):
            network.get_radial_routes("bus_obj")

    def test_get_diametral_routes(self):
        """
        to test add diametral routes
        :return:
        """
        g = graph.Graph.build_from_parameters(5, 1000, 0.5, 2)
        m = TransportModeManager()
        bus_obj = m.get_mode("bus")
        network = TransportNetwork(g)

        routes = network.get_diametral_routes(mode_obj=bus_obj, jump=2, short=False, express=False)

        self.assertEqual(len(routes), 5)
        self.assertEqual(routes[0].nodes_sequence_i, [1, 2, 0, 6, 5])
        self.assertEqual(routes[0].nodes_sequence_r, [5, 6, 0, 2, 1])
        self.assertEqual(routes[0].stops_sequence_i, [1, 2, 0, 6, 5])
        self.assertEqual(routes[0].stops_sequence_r, [5, 6, 0, 2, 1])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_i, [9, 10, 0, 4, 3])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_r, [3, 4, 0, 10, 9])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_i, [9, 10, 0, 4, 3])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_r, [3, 4, 0, 10, 9])

        routes = network.get_diametral_routes(mode_obj=bus_obj, jump=2, short=True, express=False)

        self.assertEqual(len(routes), 5)
        self.assertEqual(routes[0].nodes_sequence_i, [2, 0, 6])
        self.assertEqual(routes[0].nodes_sequence_r, [6, 0, 2])
        self.assertEqual(routes[0].stops_sequence_i, [2, 0, 6])
        self.assertEqual(routes[0].stops_sequence_r, [6, 0, 2])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_i, [10, 0, 4])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_r, [4, 0, 10])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_i, [10, 0, 4])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_r, [4, 0, 10])

        routes = network.get_diametral_routes(mode_obj=bus_obj, jump=2, short=False, express=True)

        self.assertEqual(len(routes), 5)
        self.assertEqual(routes[0].nodes_sequence_i, [1, 2, 0, 6, 5])
        self.assertEqual(routes[0].nodes_sequence_r, [5, 6, 0, 2, 1])
        self.assertEqual(routes[0].stops_sequence_i, [1, 5])
        self.assertEqual(routes[0].stops_sequence_r, [5, 1])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_i, [9, 10, 0, 4, 3])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_r, [3, 4, 0, 10, 9])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_i, [9, 3])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_r, [3, 9])

        routes = network.get_diametral_routes(mode_obj=bus_obj, jump=2, short=True, express=True)

        self.assertEqual(len(routes), 5)
        self.assertEqual(routes[0].nodes_sequence_i, [2, 0, 6])
        self.assertEqual(routes[0].nodes_sequence_r, [6, 0, 2])
        self.assertEqual(routes[0].stops_sequence_i, [2, 6])
        self.assertEqual(routes[0].stops_sequence_r, [6, 2])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_i, [10, 0, 4])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_r, [4, 0, 10])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_i, [10, 4])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_r, [4, 10])

        routes = network.get_diametral_routes(mode_obj=bus_obj, jump=1, short=True, express=True)

        self.assertEqual(len(routes), 5)
        self.assertEqual(routes[0].nodes_sequence_i, [2, 0, 4])
        self.assertEqual(routes[0].nodes_sequence_r, [4, 0, 2])
        self.assertEqual(routes[0].stops_sequence_i, [2, 4])
        self.assertEqual(routes[0].stops_sequence_r, [4, 2])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_i, [10, 0, 2])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_r, [2, 0, 10])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_i, [10, 2])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_r, [2, 10])

        with self.assertRaises(exceptions.JumpIsNotValidException):
            network.get_diametral_routes(mode_obj=bus_obj, jump=6, short=True, express=True)

        with self.assertRaises(exceptions.ModeIsNotValidException):
            network.get_diametral_routes(mode_obj="bus_obj", jump=6, short=True, express=True)

    def test_get_tangencial_routes(self):
        """
        to test add tangencial routes
        :return:
        """
        g = graph.Graph.build_from_parameters(5, 1000, 0.5, 2)
        m = TransportModeManager()
        bus = m.get_mode("bus")
        t = TransportNetwork(g)

        routes = t.get_tangencial_routes(mode_obj=bus, jump=2, short=False, express=False)

        self.assertEqual(len(routes), 5)
        self.assertEqual(routes[0].nodes_sequence_i, [1, 2, 4, 6, 5])
        self.assertEqual(routes[0].nodes_sequence_r, [5, 6, 4, 2, 1])
        self.assertEqual(routes[0].stops_sequence_i, [1, 2, 4, 6, 5])
        self.assertEqual(routes[0].stops_sequence_r, [5, 6, 4, 2, 1])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_i, [9, 10, 2, 4, 3])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_r, [3, 4, 2, 10, 9])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_i, [9, 10, 2, 4, 3])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_r, [3, 4, 2, 10, 9])

        routes = t.get_tangencial_routes(mode_obj=bus, jump=2, short=True, express=False)

        self.assertEqual(len(routes), 5)
        self.assertEqual(routes[0].nodes_sequence_i, [2, 4, 6])
        self.assertEqual(routes[0].nodes_sequence_r, [6, 4, 2])
        self.assertEqual(routes[0].stops_sequence_i, [2, 4, 6])
        self.assertEqual(routes[0].stops_sequence_r, [6, 4, 2])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_i, [10, 2, 4])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_r, [4, 2, 10])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_i, [10, 2, 4])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_r, [4, 2, 10])

        routes = t.get_tangencial_routes(mode_obj=bus, jump=2, short=False, express=True)

        self.assertEqual(len(routes), 5)
        self.assertEqual(routes[0].nodes_sequence_i, [1, 2, 4, 6, 5])
        self.assertEqual(routes[0].nodes_sequence_r, [5, 6, 4, 2, 1])
        self.assertEqual(routes[0].stops_sequence_i, [1, 5])
        self.assertEqual(routes[0].stops_sequence_r, [5, 1])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_i, [9, 10, 2, 4, 3])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_r, [3, 4, 2, 10, 9])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_i, [9, 3])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_r, [3, 9])

        routes = t.get_tangencial_routes(mode_obj=bus, jump=2, short=True, express=True)

        self.assertEqual(len(routes), 5)
        self.assertEqual(routes[0].nodes_sequence_i, [2, 4, 6])
        self.assertEqual(routes[0].nodes_sequence_r, [6, 4, 2])
        self.assertEqual(routes[0].stops_sequence_i, [2, 6])
        self.assertEqual(routes[0].stops_sequence_r, [6, 2])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_i, [10, 2, 4])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_r, [4, 2, 10])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_i, [10, 4])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_r, [4, 10])

        routes = t.get_tangencial_routes(mode_obj=bus, jump=1, short=True, express=True)

        self.assertEqual(len(routes), 5)
        self.assertEqual(routes[0].nodes_sequence_i, [2, 4])
        self.assertEqual(routes[0].nodes_sequence_r, [4, 2])
        self.assertEqual(routes[0].stops_sequence_i, [2, 4])
        self.assertEqual(routes[0].stops_sequence_r, [4, 2])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_i, [10, 2])
        self.assertEqual(routes[len(routes) - 1].nodes_sequence_r, [2, 10])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_i, [10, 2])
        self.assertEqual(routes[len(routes) - 1].stops_sequence_r, [2, 10])

        with self.assertRaises(exceptions.JumpIsNotValidException):
            t.get_tangencial_routes(mode_obj=bus, jump=6, short=True, express=True)

        with self.assertRaises(exceptions.ModeIsNotValidException):
            t.get_tangencial_routes(mode_obj="bus", jump=6, short=True, express=True)

    # def test_plot(self):
    #     """
    #     to test plot method
    #     :return:
    #     """
    #     g = graph.Graph.build_from_parameters(7, 1000, 0.5, 0,
    #                                           angles=[10, 50, 150, 180, 270,
    #                                                   300, 320], etha=0.5,
    #                                           etha_zone=3,
    #                                           Hi=[1, 2, 1, 1, 1, 0.5, 3],
    #                                           Gi=[1, 2, 1, 1, 1, 3, 2])
    #
    #     t = TransportNetwork(g)
    #     m = TransportModeManager()
    #     bus = m.get_mode("bus")
    #     routes = t.get_radial_routes(bus)
    #
    #     for route in routes:
    #         t.add_route(route)
    #
    #     # save figure in path
    #     t.plot(os.path.join(self.data_path, 'figure2_test.png'))
    #     file_obj = Path(os.path.join(self.data_path, 'figure2_test.png'))
    #     # test
    #     self.assertTrue(file_obj.is_file())
    #     # to compare figure with a test figure
    #     self.assertTrue(
    #         filecmp.cmp(os.path.join(self.data_path, 'figure1_test.png'),
    #                     os.path.join(self.data_path, "figure2_test.png")))
    #
    #     # remove file
    #     os.remove(os.path.join(self.data_path, 'figure2_test.png'))
    #
    #     with self.assertRaises(exceptions.RouteIdNotFoundException):
    #         t.plot(os.path.join(self.data_path, 'figure_test2.png'),
    #                list_routes=["506"])
    #         # remove file
    #         os.remove(os.path.join(self.data_path, 'figure2_test.png'))



import filecmp
import os
import unittest
from pathlib import Path

from sidermit.city import demand, graph
from sidermit import exceptions


class test_graph(unittest.TestCase):

    def setUp(self):
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        self.data_path = os.path.join(self.dir_path, 'file')
        self.data_path = os.path.join(self.data_path, 'demand')

    def test_symmetric_parameters_exceptions(self):
        """
        to test symmetric parameters exceptions y, a, alpha, beta
        :return:
        """

        g = graph.Graph.build_from_parameters(3, 1000, 0.5, 2)

        with self.assertRaises(exceptions.YIsNotValidException):
            demand.Demand.build_from_parameters(g, None, 0.5, 1 / 3, 1 / 3)
        with self.assertRaises(exceptions.AIsNotValidException):
            demand.Demand.build_from_parameters(g, 1000, None, 1 / 3, 1 / 3)
        with self.assertRaises(exceptions.AlphaIsNotValidException):
            demand.Demand.build_from_parameters(g, 1000, 0.5, None, 1 / 3)
        with self.assertRaises(exceptions.BetaIsNotValidException):
            demand.Demand.build_from_parameters(g, 1000, 0.5, 1 / 3, None)

        with self.assertRaises(exceptions.YOutOfRangeException):
            demand.Demand.build_from_parameters(g, -100, 0.5, 1 / 3, 1 / 3)
        with self.assertRaises(exceptions.AOutOfRangeException):
            demand.Demand.build_from_parameters(g, 1000, 2, 1 / 3, 1 / 3)
        with self.assertRaises(exceptions.AlphaOutOfRangeException):
            demand.Demand.build_from_parameters(g, 1000, 0.5, 2, 1 / 3)
        with self.assertRaises(exceptions.BetaOutOfRangeException):
            demand.Demand.build_from_parameters(g, 1000, 0.5, 1 / 3, 2)
        with self.assertRaises(exceptions.AlphaBetaOutOfRangeException):
            demand.Demand.build_from_parameters(g, 1000, 0.5, 0.8, 0.8)

    def test_build_from_parameters(self):
        """
        to test build_from_parameters method
        :return:
        """
        g = graph.Graph.build_from_parameters(0, 1000, 0.5, 2)

        d = demand.Demand.build_from_parameters(g, 1000, 0.5, 1 / 3, 1 / 3)

        self.assertEqual(len(d.get_matrix()), 1)

        g = graph.Graph.build_from_parameters(1, 1000, 0.5, 2)

        d = demand.Demand.build_from_parameters(g, 1000, 0.5, 1 / 3, 1 / 3)

        self.assertEqual(len(d.get_matrix()), 3)

        g = graph.Graph.build_from_parameters(2, 1000, 0.5, 2)

        d = demand.Demand.build_from_parameters(g, 1000, 0.5, 1 / 3, 1 / 3)

        self.assertEqual(len(d.get_matrix()), 5)

    def test_build_from_file(self):
        """
        to test build_from_file method
        :return:
        """
        g = graph.Graph.build_from_parameters(3, 1000, 0.5, 2)
        d = demand.Demand.build_from_file(g, os.path.join(self.data_path, 'test_matrix.csv'))

        self.assertEqual(len(d.get_matrix()), 7)

        with self.assertRaises(exceptions.NodeIdDemandIsNotAnInteger):
            demand.Demand.build_from_file(g, os.path.join(self.data_path, 'test_stringid.csv'))

    def test_matrix_to_file(self):
        """
        to test matrix_to_file method
        :return:
        """

        # build graph
        g = graph.Graph.build_from_parameters(2, 1000, 0.5, 2)
        d = demand.Demand.build_from_parameters(g, 1000, 0.5, 1 / 3, 1 / 3)
        # write file
        d.matrix_to_file(os.path.join(self.data_path, 'write_test.csv'))
        fileObj = Path(os.path.join(self.data_path, 'write_test.csv'))
        # test
        self.assertTrue(fileObj.is_file())
        # to compare file with a test file
        self.assertTrue(
            filecmp.cmp(os.path.join(self.data_path, 'write1_test.csv'),
                        os.path.join(self.data_path, 'write_test.csv')))
        # remove file
        os.remove(os.path.join(self.data_path, 'write_test.csv'))

    def test_change_vij_exceptions(self):
        """
        to test exceptions of change_vij method
        :return:
        """

        with self.assertRaises(exceptions.TripsValueIsNotValidException):
            g = graph.Graph.build_from_parameters(3, 1000, 0.5, 2)
            demand.Demand.build_from_file(g, os.path.join(self.data_path, 'test_raisesVijNegativeExceptions.csv'))

        with self.assertRaises(exceptions.DestinationIdDoesNotFoundException):
            g = graph.Graph.build_from_parameters(3, 1000, 0.5, 2)
            demand.Demand.build_from_file(g,
                                          os.path.join(self.data_path, 'test_raisesDestinationNotFoundExceptions.csv'))

        with self.assertRaises(exceptions.OriginIdDoesNotFoundException):
            g = graph.Graph.build_from_parameters(3, 1000, 0.5, 2)
            demand.Demand.build_from_file(g, os.path.join(self.data_path, 'test_raisesOriginNotFoundExceptions.csv'))

    def test_build_from_file_exceptions(self):
        """
        to test exceptions of build_from_file method
        :return:
        """

        with self.assertRaises(exceptions.FileFormatIsNotValidException):
            g = graph.Graph.build_from_parameters(3, 1000, 0.5, 2)
            demand.Demand.build_from_file(g, os.path.join(self.data_path, 'test_raisesFormatExceptions.csv'))


if __name__ == '__main__':
    unittest.main()



from collections import defaultdict
from typing import List

from sidermit.optimization.preoptimization import ExtendedEdge, ExtendedNode
from sidermit.optimization.preoptimization import StopNode, RouteNode
from sidermit.publictransportsystem import RouteType, Route

defaultdict2_float = defaultdict(lambda: defaultdict(float))
defaultdict3_float = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
list_suc = defaultdict(List[ExtendedEdge])
list_lab = defaultdict(float)
list_f = defaultdict(float)
list_elemental_path = List[ExtendedNode]
defaultdict_elemental_path = defaultdict(List[list_elemental_path])

dic_hyperpaths = defaultdict(lambda: defaultdict(lambda: defaultdict(List[list_elemental_path])))
dic_labels = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
dic_successors = defaultdict(lambda: defaultdict(lambda: defaultdict(List[ExtendedEdge])))
dic_frequency = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
dic_Vij = defaultdict(lambda: defaultdict(float))
dic_assigment = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
dic_f = defaultdict(float)

dic_boarding = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
dic_alighting = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))

dic_loaded_section = defaultdict(float)


class Assignment:

    def __init__(self):
        pass

    @staticmethod
    def get_assignment(hyperpaths: dic_hyperpaths, labels: dic_labels, p: float, vp: float, spa: float,
                       spv: float) -> dic_assigment:
        """
        to distribute trips of all OD pair in each StopNode of the Origin
        :param vp: Walking speed [km/h]
        :param spv: Subjetive value of in-vehicle time savings [US$/h]
        :param spa: Subjetive value of access time savings [US$/h]
        :param hyperpaths: Dic[origin: CityNode][destination: CityNode][StopNode] = List[List[ExtendedNodes]].
        Each List[ExtendedNodes] represent a elemental path.
        :param labels: dic[origin: CityNode][destination: CityNode][ExtendedNode] = Label [
        :param p: width [m] of all CityNode
        :return: dic[origin: CityNode][destination: CityNode][Stop: StopNode] = %V_OD
        """

        assignment = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))

        for origin in hyperpaths:
            for destination in hyperpaths[origin]:

                # paradero de d = 1
                stop1 = None
                # otro paradero ( su d puede ser o no 1)
                stop2 = None

                # encontramos paradero de d = 1
                for stop in hyperpaths[origin][destination]:
                    if stop.mode.d == 1:
                        if stop1 is None:
                            stop1 = stop
                        else:
                            stop2 = stop
                    else:
                        stop2 = stop

                # solo tiene una parada
                if stop1 is None or stop2 is None:
                    if stop1 is not None:
                        assignment[origin][destination][stop1] = 100
                    if stop2 is not None:
                        assignment[origin][destination][stop2] = 100
                # existen ambas paradas
                else:
                    # paradero con d = 1 es de etiqueta minima
                    if labels[origin][destination][stop1] <= labels[origin][destination][stop2]:
                        # calculamos caminata de indiferencia
                        d = vp * (labels[origin][destination][stop2] - labels[origin][destination][stop1]) / (spa / spv)

                        # caminata de indiferencia es mayor a la zona de influencia de stop1
                        if d >= p / 2:
                            assignment[origin][destination][stop1] = 100

                        # caminata de indiferencia es menor a la zona de influencia de stop1
                        else:

                            # zona de influencia de stop_2
                            zona_stop_2 = p / stop2.mode.d

                            # encontraremos linea de stop 2
                            position = 0
                            # reconoceremos posición de todos los paraderos ubicado a la derecha de stop1
                            for i in range(int(stop2.mode.d / 2)):
                                if i == 0:
                                    position = zona_stop_2 / 2
                                else:
                                    position = position + zona_stop_2
                                # encontramos un paradero de stop2 que esta ubicado mas lejos que la distancia de
                                # indiferencia
                                if position > d:
                                    assignment[origin][destination][stop1] = (2 * d + (position - d)) / p * 100
                                    assignment[origin][destination][stop2] = 100 - assignment[origin][destination][
                                        stop1]
                                    break
                            # si no se encontro paradero mas lejos a la distancia de indiferencia asignar to do a stop1
                            if position < d:
                                assignment[origin][destination][stop1] = 100
                    else:
                        # si parametro d de stop2 es impar
                        if stop2.mode.d % 2 == 1:
                            assignment[origin][destination][stop2] = 100
                        else:
                            # calculamos caminata de indiferencia
                            d = vp * (labels[origin][destination][stop1] - labels[origin][destination][stop2]) / (
                                    spa / spv)

                            # necesitamos posicion del primer paradero stop2 a la derecha del centro
                            position = (p / stop2.mode.d) * 0.5

                            if d >= position:
                                assignment[origin][destination][stop2] = 100
                            else:
                                assignment[origin][destination][stop1] = (position - d) / p * 100
                                assignment[origin][destination][stop2] = 100 - assignment[origin][destination][
                                    stop1]
        return assignment

    @staticmethod
    def get_alighting_and_boarding(Vij: dic_Vij, hyperpaths: dic_hyperpaths, successors: dic_successors,
                                   assignment: dic_assigment, f: dic_f) -> (dic_boarding, dic_alighting):
        """
        to get two matrix (z and v) with alighting and boarding for vehicle in each stop of all routes
         :param successors: dic[origin: CityNode][destination: CityNode] [ExtendedNode] = List[ExtendedEdge],
        List[ExtendedEdge] represent all successors edge for each ExtendedNode in a OD pair.
        :param Vij: dic[origin: CityNode][destination: CityNode] = vij [pax/hr]
        :param hyperpaths: Dic[origin: CityNode][destination: CityNode][StopNode] = List[List[ExtendedNodes]].
        Each List[ExtendedNodes] represent a elemental path to connect a origin and destination
        :param assignment: dic[origin: CityNode][destination: CityNode][Stop: StopNode] =%V_OD
        :param f: dic[route_id] = frequency [veh/hr]
        :return: (z,v,loaded_section_route)
        z = dic[route_id][direction][stop: StopNode] = pax [pax/veh],
        v = dic[route_id][direction][stop: StopNode] = pax [pax/veh],
        loaded_section_route = dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        """
        # dic[route_id][direction][stop: StopNode] = pax[pax / veh]
        z = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
        # dic[route_id][direction][stop: StopNode] = pax[pax / veh]
        v = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
        # dic[route_id][direction][stop: StopNode] = pax[pax / veh]
        loaded_section_route = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))

        for origin in hyperpaths:
            for destination in hyperpaths[origin]:
                # viajes del par OD
                vod = Vij[origin][destination]
                for stop in hyperpaths[origin][destination]:
                    # viajes de todas las rutas elementales que salen de esta parada
                    vod_s = vod * assignment[origin][destination][stop] / 100

                    if vod_s == 0:
                        continue

                    # cosntruye las rutas elementales del par OD que surgen de la parada
                    paths = []
                    # constituida de tuplas, noodei, nodoj, vij
                    for suc in successors[origin][destination][stop]:
                        nodej = suc.nodej
                        paths.append((stop, nodej, vod_s))

                    while len(paths) != 0:
                        nodei, nodej, pax = paths.pop(0)

                        dis_pax = pax

                        # arco de subida
                        if isinstance(nodei, StopNode):
                            if isinstance(nodej, RouteNode):
                                # cambia la distribucion de pasajeros
                                # aumentan las subidas
                                f_acum = 0

                                for suc in successors[origin][destination][nodei]:
                                    f_acum += f[suc.nodej.route.id]

                                dis_pax = pax * (f[nodej.route.id] / f_acum)

                                z[nodej.route.id][nodej.direction][nodei] += dis_pax

                        # arco de bajada
                        if isinstance(nodei, RouteNode):
                            if isinstance(nodej, StopNode):
                                # aumentan las bajadas
                                v[nodei.route.id][nodei.direction][nodej] += dis_pax
                        # arco de ruta
                        if isinstance(nodei, RouteNode):
                            if isinstance(nodej, RouteNode):
                                # aumentan las cargas por tramo
                                loaded_section_route[nodei.route.id][nodei.direction][nodei.stop_node] += dis_pax

                        # agregar nuevos elementos a paths, salvo que hayan llegado a destino
                        if isinstance(nodej, StopNode) and nodej.city_node == destination:
                            continue

                        else:
                            for suc in successors[origin][destination][nodej]:
                                paths.append((nodej, suc.nodej, dis_pax))

        for route_id in z:
            for direction in z[route_id]:
                for stop_node in z[route_id][direction]:
                    if f[route_id] == 0:
                        continue
                    else:
                        z[route_id][direction][stop_node] = z[route_id][direction][stop_node] / (
                            f[route_id])
        for route_id in v:
            for direction in v[route_id]:
                for stop_node in v[route_id][direction]:
                    if f[route_id] == 0:
                        continue
                    else:
                        v[route_id][direction][stop_node] = v[route_id][direction][stop_node] / (
                            f[route_id])

        for route_id in loaded_section_route:
            for direction in loaded_section_route[route_id]:
                for stop_node in loaded_section_route[route_id][direction]:
                    if f[route_id] == 0:
                        continue
                    else:
                        loaded_section_route[route_id][direction][stop_node] = \
                            loaded_section_route[route_id][direction][stop_node] / (f[route_id])

        return z, v, loaded_section_route

    @staticmethod
    def str_boarding_alighting(z: dic_boarding, v: dic_alighting) -> str:
        """
        to print boarding and alighting
        :param z: boarding, dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :param v: alighting, dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :return:
        """
        line = "\nBoarding and Alighting information:"
        for route_id in z:
            line += "\nNew route: {}".format(route_id)
            line += "\n\tBoarding:"
            for direction in z[route_id]:
                line += "\n\t\tDirection: {}".format(direction)
                for stop_node in z[route_id][direction]:
                    line += "\n\t\t\tStop {}-{}: {:.2f}[pax/veh]".format(stop_node.mode.name,
                                                                         stop_node.city_node.graph_node.name,
                                                                         z[route_id][direction][stop_node])
            line += "\n\tAlighting:"
            for direction in v[route_id]:
                line += "\n\t\tDirection: {}".format(direction)
                for stop_node in v[route_id][direction]:
                    line += "\n\t\t\tStop {}-{}: {:.2f}[pax/veh]".format(stop_node.mode.name,
                                                                         stop_node.city_node.graph_node.name,
                                                                         v[route_id][direction][stop_node])
        return line

    @staticmethod
    def most_loaded_section(loaded_section_route: defaultdict3_float) -> dic_loaded_section:
        """
        to get  most loaded section for each routes
        :param loaded_section_route: dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :return: dic[route_id] = pax [pax/veh]
        """

        most_loaded_section = defaultdict(float)
        for route_id in loaded_section_route:
            max_load = 0
            for direction in loaded_section_route[route_id]:
                for stop_node in loaded_section_route[route_id][direction]:
                    if loaded_section_route[route_id][direction][stop_node] > max_load:
                        max_load = loaded_section_route[route_id][direction][stop_node]

            most_loaded_section[route_id] = max_load

        return most_loaded_section

    @staticmethod
    def most_loaded_section_2(routes: List[Route], z: dic_boarding, v: dic_alighting,
                              f: list_f) -> dic_loaded_section:
        """
        to get  most loaded section for each routes
        :param f: dic[route_id] = frequency [veh/hr]
        :param routes: List[Route]
        :param z: boarding, dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :param v: alighting, dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :return: dic[route_id] = pax [pax/veh]
        """

        most_loaded_section = defaultdict(float)

        for route in routes:

            route_id = route.id
            node_sequence_i = route.nodes_sequence_i
            node_sequence_r = route.nodes_sequence_r

            # caso circular
            if route._type == RouteType.CIRCULAR:

                # circular con sentido de ida
                if len(node_sequence_i) > 0:
                    node_sequence = node_sequence_i
                    direction = "I"
                # circular con sentido de vuelta
                else:
                    node_sequence = node_sequence_r
                    direction = "R"

                qi = [0]
                for i in node_sequence:
                    zi = 0
                    vi = 0
                    for stop_node in z[route_id][direction]:
                        if str(stop_node.city_node.graph_node.id) == str(i):
                            zi = z[route_id][direction][stop_node]
                            break
                    for stop_node in v[route_id][direction]:
                        if str(stop_node.city_node.graph_node.id) == str(i):
                            vi = v[route_id][direction][stop_node]
                            break
                    new_qi = qi[len(qi) - 1] + (zi - vi) * f[route_id]
                    qi.append(new_qi)
                q = min(qi)
                ki_max = 0
                for i in range(len(qi)):
                    if i == 0:
                        continue
                    ki = (qi[i] - q) / f[route_id]
                    if ki > ki_max:
                        ki_max = ki
                most_loaded_section[route_id] = ki_max

            # caso no circular
            else:

                max_loaded_section = 0
                prev_loaded = 0

                for i in node_sequence_i:
                    zi = 0
                    vi = 0

                    for stop_node in z[route_id]["I"]:
                        if str(stop_node.city_node.graph_node.id) == str(i):
                            zi = z[route_id]["I"][stop_node]
                            break
                    for stop_node in v[route_id]["I"]:
                        if str(stop_node.city_node.graph_node.id) == str(i):
                            vi = v[route_id]["I"][stop_node]
                            break

                    prev_loaded += zi - vi

                    if prev_loaded > max_loaded_section:
                        max_loaded_section = prev_loaded

                prev_loaded = 0

                for i in node_sequence_r:

                    zi = 0
                    vi = 0

                    for stop_node in z[route_id]["R"]:
                        if str(stop_node.city_node.graph_node.id) == str(i):
                            zi = z[route_id]["R"][stop_node]
                            break
                    for stop_node in v[route_id]["R"]:
                        if str(stop_node.city_node.graph_node.id) == str(i):
                            vi = v[route_id]["R"][stop_node]
                            break

                    prev_loaded += zi - vi

                    if prev_loaded > max_loaded_section:
                        max_loaded_section = prev_loaded

                most_loaded_section[route_id] = max_loaded_section

        return most_loaded_section



from .mode import TransportMode, TransportModeManager
from .network import TransportNetwork, Route, RouteType
from .passenger import Passenger

__all__ = ['TransportMode', 'TransportModeManager', 'TransportNetwork',
           'Passenger', 'Route', 'RouteType']



import unittest

from sidermit import exceptions
from sidermit.publictransportsystem import Passenger


class PassengerTest(unittest.TestCase):

    def test_create_passenger(self):
        """

        """
        va = 1
        pv = 2
        pw = 3
        pa = 4
        pt = 5
        spv = 6
        spw = 7
        spa = 8
        spt = 9

        passenger_obj = Passenger(va, pv, pw, pa, pt, spv, spw, spa, spt)

        self.assertEqual(passenger_obj.va, va)
        self.assertEqual(passenger_obj.pv, pv)
        self.assertEqual(passenger_obj.pw, pw)
        self.assertEqual(passenger_obj.pa, pa)
        self.assertEqual(passenger_obj.pt, pt)
        self.assertEqual(passenger_obj.spv, spv)
        self.assertEqual(passenger_obj.spw, spw)
        self.assertEqual(passenger_obj.spa, spa)
        self.assertEqual(passenger_obj.spt, spt)

    def test_raises_user_exceptions(self):
        """
        test user exceptions
        :return:
        """

        with self.assertRaises(exceptions.VaIsNotValidExceptions):
            Passenger(0, 1, 1, 1, 1, 1, 1, 1, 1)
        with self.assertRaises(exceptions.PvIsNotValidExceptions):
            Passenger(1, -1, 1, 1, 1, 1, 1, 1, 1)
        with self.assertRaises(exceptions.PwIsNotValidExceptions):
            Passenger(1, 1, -1, 1, 1, 1, 1, 1, 1)
        with self.assertRaises(exceptions.PaIsNotValidExceptions):
            Passenger(1, 1, 1, -1, 1, 1, 1, 1, 1)
        with self.assertRaises(exceptions.PtIsNotValidExceptions):
            Passenger(1, 1, 1, 1, -1, 1, 1, 1, 1)
        with self.assertRaises(exceptions.SpvIsNotValidExceptions):
            Passenger(1, 1, 1, 1, 1, -1, 1, 1, 1)
        with self.assertRaises(exceptions.SpwIsNotValidExceptions):
            Passenger(1, 1, 1, 1, 1, 1, -1, 1, 1)
        with self.assertRaises(exceptions.SpaIsNotValidExceptions):
            Passenger(1, 1, 1, 1, 1, 1, 1, -1, 1)
        with self.assertRaises(exceptions.SptIsNotValidExceptions):
            Passenger(1, 1, 1, 1, 1, 1, 1, 1, -1)

    def test_get_default_passenger(self):
        passenger_obj = Passenger.get_default_passenger()

        self.assertEqual(passenger_obj.va, 4.0)
        self.assertEqual(passenger_obj.pv, 2.74)
        self.assertEqual(passenger_obj.pw, 5.48)
        self.assertEqual(passenger_obj.pa, 8.22)
        self.assertEqual(passenger_obj.pt, 16)
        self.assertEqual(passenger_obj.spv, 2.74)
        self.assertEqual(passenger_obj.spw, 5.48)
        self.assertEqual(passenger_obj.spa, 8.22)
        self.assertEqual(passenger_obj.spt, 16)



# General exception
class SIDERMITException(Exception):
    pass


# Optimizer exception
class OptimizerException(SIDERMITException):
    pass


# Graph exception
class GraphException(SIDERMITException):
    pass


# Demand exception
class DemandException(SIDERMITException):
    pass


# Transport Mode exception
class TransportModeException(SIDERMITException):
    pass


# Transport User exception
class TransportUserException(SIDERMITException):
    pass


# Transport Network Exceptions
class TransportNetworkException(SIDERMITException):
    pass


class HyperpathException(SIDERMITException):
    pass


# Graph Exceptions


# build Zone
class PeripheryNodeTypeIsNotValidException(GraphException):
    pass


class SubcenterNodeTypeIsNotValidException(GraphException):
    pass


# build Node
class NodeRadiusIsNotValidException(GraphException):
    pass


class NodeAngleIsNotValidException(GraphException):
    pass


class NodeWidthIsNotValidException(GraphException):
    pass


# build graph
class CBDDoesNotExistException(GraphException):
    pass


class ZoneIdIsNotValidException(GraphException):
    pass


class AddPreviousZonesException(GraphException):
    pass


class PeripheryTypeIsNotValidException(GraphException):
    pass


class SubcenterTypeIsNotValidException(GraphException):
    pass


class CBDTypeIsNotValidException(GraphException):
    pass


class CBDDuplicatedException(GraphException):
    pass


class NodeIdDuplicatedException(GraphException):
    pass


class EdgeDoesNotExistException(GraphException):
    pass


class EdgeIdDuplicatedException(GraphException):
    pass


class NIsNotDefined(GraphException):
    pass


class NIsNotValidNumberException(GraphException):
    pass


class LIsNotDefined(GraphException):
    pass


class LIsNotValidNumberException(GraphException):
    pass


class GIsNotDefined(GraphException):
    pass


class GIsNotValidNumberException(GraphException):
    pass


class PIsNotDefined(GraphException):
    pass


class PIsNotValidNumberException(GraphException):
    pass


class EdgeIsNotAvailableException(GraphException):
    pass


class PajekFormatIsNotValidException(GraphException):
    pass


class FileFormatIsNotValidException(GraphException):
    pass


class NodeTypeIsNotValidException(GraphException):
    pass


class EthaValueRequiredException(GraphException):
    pass


class EthaZoneValueRequiredException(GraphException):
    pass


class EthaValueIsNotValidException(GraphException):
    pass


class EthaZoneValueIsNotValidException(GraphException):
    pass


class AngleListLengthIsNotValidException(GraphException):
    pass


class AngleValueIsNotValidException(GraphException):
    pass


class GiListLengthIsNotValidException(GraphException):
    pass


class GiValueIsNotValidException(GraphException):
    pass


class HiListLengthIsNotValidException(GraphException):
    pass


class HiValueIsNotValidException(GraphException):
    pass


class LineNumberInFileIsNotValidException(GraphException):
    pass


class PeripherySubcenterNumberForZoneException(GraphException):
    pass


class PeripheryDoesNotExistException(GraphException):
    pass


class SubcenterDoesNotExistException(GraphException):
    pass


class NodeIdIsNotValidException(GraphException):
    pass


class EdgeIdIsNotValidException(GraphException):
    pass


class NodeDoesNotExistException(GraphException):
    pass


class WritePajekFileException(GraphException):
    pass


class NameIsNotDefinedException(GraphException):
    pass


class NodeIdIsNotAnInteger(GraphException):
    pass


class NodeIdDemandIsNotAnInteger(DemandException):
    pass


class VijIsNotValidException(DemandException):
    pass


class DemandMatrixIsNotValidException(DemandException):
    pass


class ZoneIdIsNotAnInteger(GraphException):
    pass


# Demand Exceptions
class GraphIsNotValidException(DemandException):
    pass


class NIsNotValidException(DemandException):
    pass


class YIsNotValidException(DemandException):
    pass


class AIsNotValidException(DemandException):
    pass


class AlphaIsNotValidException(DemandException):
    pass


class BetaIsNotValidException(DemandException):
    pass


class NOutOfRangeException(DemandException):
    pass


class YOutOfRangeException(DemandException):
    pass


class AOutOfRangeException(DemandException):
    pass


class AlphaOutOfRangeException(DemandException):
    pass


class BetaOutOfRangeException(DemandException):
    pass


class AlphaBetaOutOfRangeException(DemandException):
    pass


class DestinationIdDoesNotFoundException(DemandException):
    pass


class OriginIdDoesNotFoundException(DemandException):
    pass


class TripsValueIsNotValidException(DemandException):
    pass


# TransportModeException
class ByaIsNotValidExceptions(TransportModeException):
    pass


class C2IsNotValidExceptions(TransportModeException):
    pass


class C1IsNotValidExceptions(TransportModeException):
    pass


class KmaxIsNotValidExceptions(TransportModeException):
    pass


class TIsNotValidExceptions(TransportModeException):
    pass


class FmaxIsNotValidExceptions(TransportModeException):
    pass


class ThetaIsNotValidExceptions(TransportModeException):
    pass


class DIsNotValidExceptions(TransportModeException):
    pass


class TatIsNotValidExceptions(TransportModeException):
    pass


class VIsNotValidExceptions(TransportModeException):
    pass


class CoIsNotValidExceptions(TransportModeException):
    pass


class ModeIsNotValidException(TransportModeException):
    pass


class NameIsNotValidExceptions(TransportModeException):
    pass


class FiniIsNotValidException(TransportModeException):
    pass


class AddModeExceptions(TransportModeException):
    pass


class ModeDoesNotExistExceptions(TransportModeException):
    pass


class ModeNotFoundExceptions(TransportModeException):
    pass


# transport user exceptions
class PvIsNotValidExceptions(TransportUserException):
    pass


class PaIsNotValidExceptions(TransportUserException):
    pass


class VaIsNotValidExceptions(TransportUserException):
    pass


class PwIsNotValidExceptions(TransportUserException):
    pass


class SpvIsNotValidExceptions(TransportUserException):
    pass


class SpaIsNotValidExceptions(TransportUserException):
    pass


class PtIsNotValidExceptions(TransportUserException):
    pass


class SpwIsNotValidExceptions(TransportUserException):
    pass


class SptIsNotValidExceptions(TransportUserException):
    pass


class RouteIdIsNotValidException(TransportNetworkException):
    pass


class NodeSequenceIsNotValidException(TransportNetworkException):
    pass


class StopSequenceIsNotValidException(TransportNetworkException):
    pass


class ModeNameIsNotValidException(TransportNetworkException):
    pass


class SequencesLenException(TransportNetworkException):
    pass


class StopsSequencesException(TransportNetworkException):
    pass


class FirstStopIsNotValidException(TransportNetworkException):
    pass


class LastStopIsNotValidException(TransportNetworkException):
    pass


class NotCycleException(TransportNetworkException):
    pass


class NodeSequencesIsNotValidException(TransportNetworkException):
    pass


class RouteIdNotFoundException(TransportNetworkException):
    pass


class RouteIdDuplicatedException(TransportNetworkException):
    pass


class ModeNameNotFoundException(TransportNetworkException):
    pass


class JumpIsNotValidException(TransportNetworkException):
    pass


class BanRouteIdException(TransportNetworkException):
    pass


class CircularRouteIsNotValidException(TransportNetworkException):
    pass


class RouteIsNotvalidException(TransportNetworkException):
    pass


class TransportNetworkIsNotValidException(HyperpathException):
    pass


class ConstraintViolationException(OptimizerException):
    pass


class NegativeFrequencyException(OptimizerException):
    pass


class NoOptimalSolutionFoundException(OptimizerException):
    pass



from __future__ import annotations

from collections import defaultdict
from typing import List, Tuple

import numpy as np
from scipy.optimize import minimize, NonlinearConstraint, Bounds, OptimizeResult

from sidermit.city import Graph, Demand
from sidermit.exceptions import *
from sidermit.optimization import Constrains
from sidermit.optimization import InfrastructureCost, UsersCost, OperatorsCost
from sidermit.optimization.preoptimization import Assignment, Hyperpath, ExtendedGraph, ExtendedNode, ExtendedEdge
from sidermit.publictransportsystem import Passenger
from sidermit.publictransportsystem import TransportNetwork, RouteType

defaultdict_float = defaultdict(float)
defaultdict2_float = defaultdict(lambda: defaultdict(float))
defaultdict3_float = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
list_elemental_path = List[ExtendedNode]
defaultdict_elemental_path = defaultdict(List[list_elemental_path])

dic_hyperpaths = defaultdict(lambda: defaultdict(lambda: defaultdict(List[list_elemental_path])))
dic_labels = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
dic_successors = defaultdict(lambda: defaultdict(lambda: defaultdict(List[ExtendedEdge])))
dic_frequency = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
dic_Vij = defaultdict(lambda: defaultdict(float))
dic_assigment = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
dic_f = defaultdict(float)

dic_boarding = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
dic_alighting = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))

dic_loaded_section = defaultdict(float)

defaultdict_str = defaultdict(str)


class Optimizer:
    def __init__(self, graph_obj: Graph, demand_obj: Demand, passenger_obj: Passenger, network_obj: TransportNetwork,
                 f: defaultdict_float = None):

        # definimos ciudad
        self.graph_obj = graph_obj
        _, _, _, self.p, _, _, _, _, _ = self.graph_obj.get_parameters()
        # definimos demanda
        self.demand_obj = demand_obj
        self.total_trips = demand_obj.get_total_trips()
        # definimos pasajeros
        self.passenger_obj = passenger_obj
        self.vp = self.passenger_obj.va
        self.spa = self.passenger_obj.spa
        self.spv = self.passenger_obj.spv
        self.sPTP = self.passenger_obj.spt

        # definimos red de transporte
        self.network_obj = network_obj

        # definimos frecuencia
        self.f, self.f_opt, self.lines_position = self.f0(f)

        self.extended_graph_obj = ExtendedGraph(self.graph_obj, self.network_obj.get_routes(), self.sPTP, self.f)
        self.hyperpath_obj = Hyperpath(self.extended_graph_obj, self.passenger_obj)

        # en este punto se debería levantar exception de que la red tiene mas de dos modos defnidos
        # o que existe un par OD con viaje y sin conexion
        self.hyperpaths, self.labels, self.successors, self.frequency, self.Vij = self.hyperpath_obj.get_all_hyperpaths(
            self.demand_obj.get_matrix())

        self.assignment = Assignment.get_assignment(self.hyperpaths, self.labels, self.p, self.vp, self.spa,
                                                    self.spv)

        self.len_constrains = len(self.get_constrains(self.f_opt))
        self.len_var = len(self.f_opt)

    def f0(self, f: defaultdict_float = None) -> (defaultdict_float, List[float], defaultdict_str):
        """
        to get a relation between f as a dictionary and f_opt as a list to the optimizer
        :param f: dic[route_id] = frequency [veh/hr] for all D lines
        :return: dic[route_id] = frequency [veh/hr] for all D lines, List[frequency], dic[position] = route_id
        """
        fini = defaultdict(float)
        fopt = []
        lines_position = defaultdict(None)
        n = 0
        if f is None:
            for route in self.network_obj.get_routes():
                fini[route.id] = route.mode.fini
                fopt.append(route.mode.fini)
                lines_position[n] = route.id
                n += 1
        else:
            for route in self.network_obj.get_routes():
                fini[route.id] = f[route.id]
                fopt.append(route.mode.fini)
                lines_position[n] = route.id
                n += 1

        return fini, fopt, lines_position

    def fopt_to_f(self, fopt: List[float]) -> defaultdict_float:
        """
        to get f as a dictionary
        :param fopt: List[frequency]
        :return: dic[route_id] = frequency [veh/hr] all D lines
        """
        f = defaultdict(float)
        n = 0
        for fr in fopt:
            f[self.lines_position[n]] = fr
            n += 1
        return f

    @staticmethod
    def get_k(loaded_section_route: defaultdict3_float) -> defaultdict_float:
        """
        :param loaded_section_route: dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :return: k: dic[route_id] = vehicle capacity [pax/veh]
        """
        most_loaded_section = Assignment.most_loaded_section(loaded_section_route)
        k = defaultdict(float)
        for route_id in most_loaded_section:
            k[route_id] = most_loaded_section[route_id]
        return k

    def operators_cost(self, z: defaultdict3_float, v: defaultdict3_float, f: defaultdict_float,
                       k: defaultdict_float) -> float:
        """
        to get operators cost
        :param z: boarding, dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :param v: alighting, dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :param f: dic[route_id] = frequency [veh/hr]
        :param k: dic[route_id] = frequency [pax/veh]
        :return: float, operator cost
        """
        operators_cost_obj = OperatorsCost()

        edge_distance = self.graph_obj.get_edges_distance()
        routes = self.network_obj.get_routes()
        line_travel_time = operators_cost_obj.lines_travel_time(routes, edge_distance)

        cycle_time = operators_cost_obj.get_cycle_time(z, v, routes, line_travel_time)
        cost = operators_cost_obj.get_operators_cost(routes, cycle_time, f, k)
        return cost

    def infrastructure_cost(self, f: defaultdict_float) -> float:
        """
        to get infrastructure cost
        :param f: dic[route_id] = frequency [veh/hr]
        :return: float, infrastructure cost
        """
        infrastructure_cost_obj = InfrastructureCost()
        cost = infrastructure_cost_obj.get_infrastruture_cost(self.graph_obj, self.network_obj, f)
        return cost

    def user_cost(self, hyperpaths: dic_hyperpaths, Vij: dic_Vij, assignment: dic_assigment,
                  successors: dic_successors, extended_graph: ExtendedGraph, f: defaultdict_float,
                  z: defaultdict3_float, v: defaultdict3_float) -> float:
        """
        to get users cost
        :param hyperpaths: Dic[origin: CityNode][destination: CityNode][StopNode] = List[List[ExtendedNodes]]
        :param Vij: dic[origin: CityNode][destination: CityNode] = vij
        :param assignment: dic[origin: CityNode][destination: CityNode][Stop: StopNode] = %V_OD
        :param successors: dic[origin: CityNode][destination: CityNode][ExtendedNode] = List[ExtendedEdge]
        :param extended_graph: ExtendedGraph object
        :param f: dict with frequency [veh/hr] for each route_id
        :param z: boarding, dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :param v: alighting, dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :return: float, users cost
        """
        user_cost_obj = UsersCost()
        cost = user_cost_obj.get_users_cost(hyperpaths, Vij, assignment, successors, extended_graph, f,
                                            self.passenger_obj, z, v)
        return cost

    def constrains(self, loaded_section_route: defaultdict3_float, f: defaultdict_float) -> (
            List[float], List[float]):
        """
        to get k constrains and f constrains
        :param loaded_section_route: dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :param f: dict with frequency [veh/hr] for each route_id
        :return: (k_ineq_constrains, f_ineq_constrains)
        """

        most_loaded_section = Assignment.most_loaded_section(loaded_section_route)

        constrains_obj = Constrains()

        ineq_k = constrains_obj.most_loaded_section_constrains(self.network_obj.get_routes(), most_loaded_section)
        ineq_f = constrains_obj.fmax_constrains(self.graph_obj, self.network_obj.get_routes(),
                                                self.network_obj.get_modes(), f)

        return ineq_k, ineq_f

    def VRC(self, fopt: List[float]) -> float:
        """
        to get VRC objective function to minime in optimizer
        :param fopt: variable to optimize
        :return: float, VRC value function
        """

        f = self.fopt_to_f(fopt)

        z, v, loaded_section_route = Assignment.get_alighting_and_boarding(self.Vij, self.hyperpaths, self.successors,
                                                                           self.assignment, f)
        k = self.get_k(loaded_section_route)

        CO = self.operators_cost(z, v, f, k)
        CI = self.infrastructure_cost(f)
        CU = self.user_cost(self.hyperpaths, self.Vij, self.assignment, self.successors, self.extended_graph_obj, f, z,
                            v)

        # print("f: {}".format(fopt))
        # print("VRC: {}".format(CO + CI + CU))
        return CO + CI + CU

    def get_constrains(self, fopt: List[float]) -> List[float]:
        """
        to get all constrains as a List[float]
        :param fopt: variable to optimize
        :return: all constrains as a List[float]
        """

        f = self.fopt_to_f(fopt)

        z, v, loaded_section_route = Assignment.get_alighting_and_boarding(self.Vij, self.hyperpaths, self.successors,
                                                                           self.assignment, f)

        most_loaded_section = Assignment.most_loaded_section(loaded_section_route)
        constrain_obj = Constrains()
        ineq_k = constrain_obj.most_loaded_section_constrains(self.network_obj.get_routes(), most_loaded_section)
        ineq_f = constrain_obj.fmax_constrains(self.graph_obj, self.network_obj.get_routes(),
                                               self.network_obj.get_modes(), f)

        con = []
        for c in ineq_k:
            con.append(c)
        for c in ineq_f:
            con.append(c)

        return con

    def internal_optimization(self) -> OptimizeResult:
        """
        method to do internal optimization process, with a hyperpath setted you can get a optimization of the network
        :return:     res : OptimizeResult
        The optimization result represented as a ``OptimizeResult`` object.
        Important attributes are: ``x`` the solution array, ``success`` a
        Boolean flag indicating if the optimizer exited successfully and
        ``message`` which describes the cause of the termination. See
        `OptimizeResult` for a description of other attributes.
        """

        constr_func = lambda fopt: np.array(self.get_constrains(fopt))

        lb = [-1 * np.inf] * self.len_constrains
        ub = [0] * self.len_constrains
        nonlin_con = NonlinearConstraint(constr_func, lb=lb, ub=ub)

        lb = [0] * self.len_var
        ub = [np.inf] * self.len_var

        bounds = Bounds(lb=lb, ub=ub)
        res = minimize(self.VRC, self.f_opt, method='trust-constr', constraints=nonlin_con, tol=0.01, bounds=bounds)
        print(self.print_information_internal_optimization(res))

        return res

    @staticmethod
    def print_information_internal_optimization(res: OptimizeResult) -> str:
        """
        to get a string with information about internal optimization
        :param res: OptimizeResult
        :return: str
        """

        success = res.success
        status = res.status
        message = res.message
        new_f = res.x
        constr_violation = res.constr_violation
        fun = res.fun

        line = "Internal optimization\n\tSuccess: {}\n\tStatus: {}\n\tMessage: {}\n\tnew_f: {}\n\tConstrain violation: {}\n\tVRC: {}".format(
            success, status, message, new_f, constr_violation, fun)
        return line

    @staticmethod
    def f_distance(prev_f: List[float], new_f: List[float]) -> float:
        """
        to get distance between 2 list of frequency results.
        :param prev_f: previous frequency
        :param new_f: new frequency
        :return: float, distance with normal distance
        """
        dif = 0
        for i in range(len(prev_f)):
            dif += (prev_f[i] - new_f[i]) ** len(prev_f)
        dif = dif ** (1 / len(prev_f))
        print("f_norm_distance: {}".format(dif))
        return dif

    def external_optimization_tolerance(self, prev_f: List[float], new_f: List[float], tol: float = 0.01) -> bool:
        """
        True, if tolerance criteria is success
        :param prev_f: previous frequency
        :param new_f: new frequency
        :param tol: float, tolerance
        :return: True, if tolerance criteria is success
        """

        if tol > abs(self.f_distance(prev_f, new_f)):
            return True

        return False

    @staticmethod
    def status_optimization(list_res) -> bool:
        """
        to get a information about status of external optimization
        :param list_res:List[(fopt, success, status, message, constr_violation, vrc)]
        :return: true if status is success, exceptions if not
        """
        fopt, success, status, message, constr_violation, fun = list_res[len(list_res) - 1]

        for f in fopt:
            if f < -0.1:
                raise NegativeFrequencyException("Solution with negative frequency. You can try with other fini")

        if abs(constr_violation) > 0.1:
            raise ConstraintViolationException(
                "Maximum constraint violation at the solution {}. You can try with other fini".format(constr_violation))

        if status == 0 or status == 3:
            raise NoOptimalSolutionFoundException("No optimal solution was found. You can try with other fini")

        return True

    @staticmethod
    def external_optimization(graph_obj: Graph, demand_obj: Demand, passenger_obj: Passenger,
                              network_obj: TransportNetwork,
                              f: defaultdict_float = None, tolerance: float = 0.01) -> List[Tuple]:
        """
        method to do external optimization process, several iterations of internal optimization with fixed
        hyperpaths in each
        :param graph_obj: Graph object
        :param demand_obj: Demand object
        :param passenger_obj: Passenger object
        :param network_obj: TransportNetwork object
        :param f: dict with frequency [veh/hr] for each route_id, dic[route_id] = frequency
        :param tolerance: float, tolerance to external optimization
        :return: List[(fopt, success, status, message, constr_violation, vrc)]
        """

        list_res = []

        opt_obj = Optimizer(graph_obj, demand_obj, passenger_obj, network_obj, f)
        # inicialización
        list_res.append((opt_obj.f_opt, "initialization", -1, "initialization", -1, -1))

        # primera iteracion
        res = opt_obj.internal_optimization()
        list_res.append((res.x, res.success, res.status, res.message, res.constr_violation, res.fun))

        pre_f, _, _, _, _, _ = list_res[0]
        new_f, _, _, _, _, _ = list_res[1]

        # mientras criterio de tolerancia externo se cumpla
        while not opt_obj.external_optimization_tolerance(pre_f, new_f, tolerance):
            pre_f = new_f
            dic_new_f = opt_obj.fopt_to_f(new_f)
            opt_obj = Optimizer(graph_obj, demand_obj, passenger_obj, network_obj, dic_new_f)
            res = opt_obj.internal_optimization()
            list_res.append((res.x, res.success, res.status, res.message, res.constr_violation, res.fun))
            new_f = res.x

        if opt_obj.status_optimization(list_res):
            return list_res

    @staticmethod
    def network_optimization(graph_obj: Graph, demand_obj: Demand, passenger_obj: Passenger,
                             network_obj: TransportNetwork,
                             f: defaultdict_float = None, tolerance: float = 0.01) -> Tuple:
        """
        obtain optimal frequency for the defined network if possible or raise exceptions in case of not being able
        :param graph_obj: Graph object
        :param demand_obj: Demand object
        :param passenger_obj: Passenger object
        :param network_obj: TransportNetwork object
        :param f: dict with frequency [veh/hr] for each route_id, dic[route_id] = frequency
        :param tolerance: float, tolerance to external optimization
        :return: fopt, success, status, message, constr_violation, vrc
        """

        opt_obj = Optimizer(graph_obj, demand_obj, passenger_obj, network_obj, f)
        list_res = opt_obj.external_optimization(graph_obj, demand_obj, passenger_obj, network_obj, f, tolerance)

        print(opt_obj.string_network_optimization(list_res[len(list_res) - 1]))

        return list_res[len(list_res) - 1]

    def string_network_optimization(self, res: Tuple) -> str:
        """
        get a str summary about last external optimization in network_optimization
        :param res: Tuple, (fopt, success, status, message, constr_violation, vrc)
        :return:
        """
        fopt, success, status, message, constr_violation, vrc = res
        f = self.fopt_to_f(fopt)

        line = "\n\nOptimization Results"
        line += "\nSuccess: {}".format(success)
        line += "\nStatus: {}".format(status)
        line += "\nMessage: {}".format(message)
        line += "\nMax constrain violation: {}".format(constr_violation)
        line += "\nVRC: {}".format(vrc)
        line += "\n\nFrequency information [veh/hr]: "
        for route_id in f:
            line += "\n\t{}: {:.2f}".format(route_id, f[route_id])

        return line

    def last_iteration(self, res: Tuple):
        """
        return last network optimized with optimization result
        :param res: res: Tuple, (fopt, success, status, message, constr_violation, vrc)
        :return: Optimizer object, boarding dictionary, alighting dictionary, k dictionary, loaded_section_route
        """
        fopt, success, status, message, constr_violation, vrc = res
        f = self.fopt_to_f(fopt)
        final_optimizer = Optimizer(self.graph_obj, self.demand_obj, self.passenger_obj, self.network_obj, f)
        z, v, loaded_section_route = Assignment.get_alighting_and_boarding(final_optimizer.Vij,
                                                                           final_optimizer.hyperpaths,
                                                                           final_optimizer.successors,
                                                                           final_optimizer.assignment, f)
        k = self.get_k(loaded_section_route)
        return final_optimizer, z, v, k, loaded_section_route

    def network_results(self, res: Tuple) -> List[Tuple]:
        """
        to get transport network results per line-direction
        :param res: optimization result: Tuple, (fopt, success, status, message, constr_violation, vrc)
        :return: List[(route.id: str, f [veh/hr]: float, k [pax/veh]: float, B [veh]: float, cycle_time [hr]: float,
        CO [US$/hr-pax]: float, lambda_min, sub_table_i: List[(node_i, node_j, lambda)],
        sub_table_i: List[(node_i, node_j, lambda)]
        """
        output = []

        fopt, success, status, message, constr_violation, vrc = res
        final_optimizer, z, v, k, loaded_section_route = self.last_iteration(res)
        f = self.fopt_to_f(fopt)

        # resultados de modos
        travel_time_line = OperatorsCost.lines_travel_time(final_optimizer.network_obj.get_routes(),
                                                           final_optimizer.graph_obj.get_edges_distance())
        cycle_time_line = OperatorsCost.get_cycle_time(z, v, final_optimizer.network_obj.get_routes(), travel_time_line)

        for route in final_optimizer.network_obj.get_routes():

            # flota de buses
            b = cycle_time_line[route.id] * f[route.id]
            nodes_sequence_i = route.nodes_sequence_i
            nodes_sequence_r = route.nodes_sequence_r
            # carga que hay en los tramos
            total_pax = 0
            charge_i = []
            charge_r = []
            # total de subidas a la ruta
            total_b = 0

            # caso circular
            if route._type == RouteType.CIRCULAR:

                # circular con sentido de ida
                if len(nodes_sequence_i) > 0:
                    node_sequence = nodes_sequence_i
                    direction = "I"
                # circular con sentido de vuelta
                else:
                    node_sequence = nodes_sequence_r
                    direction = "R"

                load_i = []
                for i in node_sequence:
                    for stop_node in z[route.id][direction]:
                        if stop_node.city_node.graph_node.id == i:
                            total_b += z[route.id][direction][stop_node]
                            break
                    for stop_node in loaded_section_route[route.id][direction]:
                        if stop_node.city_node.graph_node.id == i:
                            load_i.append(loaded_section_route[route.id][direction][stop_node])
                            break

                co = (route.mode.co + route.mode.c1 * k[route.id]) * f[route.id] * cycle_time_line[route.id] / (
                        total_b * f[route.id])

                charge_min = min(load_i) / max(load_i)

                sub_table = []
                sub_table_i = []
                sub_table_r = []

                node_i = None
                charge_ij = None
                for i in range(len(node_sequence)):
                    if i == 0:
                        node_i = node_sequence[i]
                        charge_ij = load_i[i] / max(load_i)
                        continue
                    else:
                        node_j = node_sequence[i]
                        sub_table.append((node_i, node_j, charge_ij))
                        node_i = node_j
                        charge_ij = load_i[i] / max(load_i)

                # circular con sentido de ida
                if len(nodes_sequence_i) > 0:
                    sub_table_i = sub_table
                # circular con sentido de vuelta
                else:
                    sub_table_r = sub_table

                output.append((
                    route.id, f[route.id], k[route.id], b, cycle_time_line[route.id] * 60, co, charge_min, sub_table_i,
                    sub_table_r))
            else:
                # z and v: dic[route_id][direction][stop: StopNode] = pax [pax/veh]
                for node_id in nodes_sequence_i:
                    bool_add = False
                    for stopnode in z[route.id]["I"]:
                        if stopnode.city_node.graph_node.id == node_id:
                            pax_b = z[route.id]["I"][stopnode]
                            pax_a = v[route.id]["I"][stopnode]
                            total_pax += pax_b - pax_a
                            total_b += pax_b
                            # print(str(node_id), total_pax, k[route.id], route.id)
                            charge_i.append((node_id, total_pax / k[route.id]))
                            bool_add = True
                            break
                    if bool_add is False:
                        charge_i.append((node_id, total_pax / k[route.id]))

                for node_id in nodes_sequence_r:
                    bool_add = False
                    for stopnode in z[route.id]["R"]:
                        if stopnode.city_node.graph_node.id == node_id:
                            pax_b = z[route.id]["R"][stopnode]
                            pax_a = v[route.id]["R"][stopnode]
                            total_pax += pax_b - pax_a
                            total_b += pax_b
                            # print(str(node_id), total_pax, k[route.id], route.id)
                            charge_r.append((node_id, total_pax / k[route.id]))
                            bool_add = True
                            break
                    if bool_add is False:
                        charge_r.append((node_id, total_pax / k[route.id]))

                co = (route.mode.co + route.mode.c1 * k[route.id]) * f[route.id] * cycle_time_line[route.id] / (
                        total_b * f[route.id])

                charge_min = float("inf")
                sub_table_i = []
                node_i = None
                charge_ij = None
                for node_j, charge in charge_i:
                    if node_i is None:
                        node_i = node_j
                        charge_ij = charge
                        continue
                    else:
                        sub_table_i.append((node_i, node_j, charge_ij))
                        if charge_min > charge_ij:
                            charge_min = charge_ij
                        node_i = node_j
                        charge_ij = charge

                sub_table_r = []
                node_i = None
                charge_ij = None
                for node_j, charge in charge_r:
                    if node_i is None:
                        node_i = node_j
                        charge_ij = charge
                        continue
                    else:
                        sub_table_r.append((node_i, node_j, charge_ij))
                        if charge_min > charge_ij:
                            charge_min = charge_ij
                        node_i = node_j
                        charge_ij = charge

                output.append((
                    route.id, f[route.id], k[route.id], b, cycle_time_line[route.id] * 60, co, charge_min, sub_table_i,
                    sub_table_r))

        return output

    @staticmethod
    def string_network_results(output_network_results: List[Tuple]) -> str:
        """
        to get a string with network results
        :param output_network_results: List[(route.id: str, f [veh/hr]: float, k [pax/veh]: float, B [veh]: float,
        cycle_time [hr]: float, CO [US$/hr-pax]: float, lambda_min, sub_table_i: List[(node_i, node_j, lambda)],
        sub_table_i: List[(node_i, node_j, lambda)]
        :return: str
        """
        line = "route_id;f[veh/hr];k[pax/veh];B[veh];tc[min];CO[US$/hr-pax];load_min;sub_table_i;sub_table_r"

        for route_id, f, k, b, cycle_time, co, charge_min, sub_table_i, sub_table_r in output_network_results:
            line += "\n{};{:.2f};{:.2f};{:.2f};{:.2f};{:.2f};{:.2f};{};{}".format(route_id, f, k, b, cycle_time, co,
                                                                                  charge_min, sub_table_i, sub_table_r)
        return line

    def write_file_network_results(self, file_path, output_network_results: List[Tuple]) -> None:
        """
        to write output file with result of optimization transport network
        :param file_path: file path
        :param output_network_results: List[(route.id: str, f [veh/hr]: float, k [pax/veh]: float, B [veh]: float,
        cycle_time [hr]: float, CO [US$/hr-pax]: float, lambda_min, sub_table_i: List[(node_i, node_j, lambda)],
        sub_table_i: List[(node_i, node_j, lambda)]
        :return: None
        """
        string_lines = self.string_network_results(output_network_results)
        file = open(file_path, 'w', encoding='utf-8')
        file.write(string_lines)
        file.close()

    def overall_results(self, res: Tuple) -> defaultdict:
        """
        to get overall cost results per passenger
        :param res: optimization result: Tuple, (fopt, success, status, message, constr_violation, vrc)
        :return: defaultdict:

        Key [unit]: value type

        VRC [USD$/hr-pax]: float,
        operators_cost [USD$/hr-pax]: float,
        infrastructure_cost [USD$/hr-pax]: float,
        users_cost [USD$/hr-pax]: float,
        travel_time_on_board [min/pax]: float,
        waiting time [min/pax]: float,
        access_time [min/pax]: float,
        transfers [transfer/pax]: float,
        vehicles_mode [veh/mode]: dic[TransportMode] = float [veh],
        vehicle_capacity_mode [pax/veh]: dic[TransportMode] = float [pax/veh],
        lines_mode : dic[TransportMode]=int [lines])
        """
        fopt, success, status, message, constr_violation, vrc = res
        final_optimizer, z, v, k, loaded_section_route = self.last_iteration(res)

        f = self.fopt_to_f(fopt)

        # resultados de costos
        CO = self.operators_cost(z, v, f, k)
        CI = self.infrastructure_cost(f)
        CU = self.user_cost(final_optimizer.hyperpaths, final_optimizer.Vij, final_optimizer.assignment,
                            final_optimizer.successors, final_optimizer.extended_graph_obj, f, z, v)
        VRC = CO + CI + CU

        # resultados de usuarios
        ta, te, tv, t = UsersCost.resources_consumer(final_optimizer.hyperpaths, final_optimizer.Vij,
                                                     final_optimizer.assignment, final_optimizer.successors,
                                                     final_optimizer.extended_graph_obj,
                                                     final_optimizer.passenger_obj.va, f, z, v)

        # resultados de modos
        travel_time_line = OperatorsCost.lines_travel_time(final_optimizer.network_obj.get_routes(),
                                                           final_optimizer.graph_obj.get_edges_distance())
        cycle_time_line = OperatorsCost.get_cycle_time(z, v, final_optimizer.network_obj.get_routes(), travel_time_line)

        B = defaultdict(float)
        L = defaultdict(int)

        K_list = defaultdict(list)
        K = defaultdict(float)

        for route in self.network_obj.get_routes():
            if f[route.id] > 0:
                B[route.mode] += f[route.id] * cycle_time_line[route.id]
                L[route.mode] += route.mode.d
                K_list[route.mode].append(k[route.id])

        for mode in K_list:
            l = K_list[mode]
            K[mode] = sum(l) / len(l)

        output = defaultdict(None)
        output["VRC"] = VRC / self.total_trips
        output["operators_cost"] = CO / self.total_trips
        output["infrastructure_cost"] = CI / self.total_trips
        output["users_cost"] = CU / self.total_trips
        output["travel_time_on_board"] = tv / self.total_trips * 60
        output["waiting time"] = te / self.total_trips * 60
        output["access_time"] = ta / self.total_trips * 60
        output["transfers"] = t / self.total_trips
        output["vehicles_mode"] = B
        output["vehicle_capacity_mode"] = K
        output["lines_mode"] = L
        return output

    @staticmethod
    def string_overall_results(overall_results: defaultdict) -> str:
        """
        to get a string to print overall results in console
        :param overall_results: defaultdict

        Key [unit]: value type

        VRC [USD$/hr-pax]: float,
        operators_cost [USD$/hr-pax]: float,
        infrastructure_cost [USD$/hr-pax]: float,
        users_cost [USD$/hr-pax]: float,
        travel_time_on_board [min/pax]: float,
        waiting time [min/pax]: float,
        access_time [min/pax]: float,
        transfers [transfer/pax]: float,
        vehicles_mode [veh/mode]: dic[TransportMode] = float [veh],
        vehicle_capacity_mode [pax/veh]: dic[TransportMode] = float [pax/veh],
        lines_mode : dic[TransportMode]=int [lines])
        :return: string to print overall results in console
        """
        vrc = overall_results["VRC"]
        co = overall_results["operators_cost"]
        ci = overall_results["infrastructure_cost"]
        cu = overall_results["users_cost"]
        tv = overall_results["travel_time_on_board"]
        te = overall_results["waiting time"]
        ta = overall_results["access_time"]
        t = overall_results["transfers"]
        b = overall_results["vehicles_mode"]
        k = overall_results["vehicle_capacity_mode"]
        l = overall_results["lines_mode"]

        line = "\n\nObjective function VRC [USD$/hr-pax]: {:.2f}".format(vrc)
        line += "\nOperators cost [USD$/hr-pax]        : {:.2f}".format(co)
        line += "\nInfrastructure cost [USD$/hr-pax]   : {:.2f}".format(ci)
        line += "\nUsers cost [USD$/hr-pax]            : {:.2f}".format(cu)

        line += "\n\nResources consumer for users"
        line += "\nTime on board vehicle [min/pax]: {:.2f}".format(tv)
        line += "\nWaiting time [min/pax]         : {:.2f}".format(te)
        line += "\nAccess time [min/pax]          : {:.2f}".format(ta)
        line += "\ntotal travel time [min/pax]    : {:.2f}".format(tv + ta + te)
        line += "\nTransfers [transfers/pax]      : {:.2f}".format(t)

        line += "\n\n"
        line += "Transport mode information: {}".format(len(b))

        for mode in b:
            line += "\n\n\tMode name: {}".format(mode.name)
            line += "\n\tB [veh]      : {:.2f}".format(b[mode])
            line += "\n\tK [pax/veh]  : {:.2f}".format(k[mode])
            line += "\n\tL [lines]    : {}".format(l[mode])

        return line



from .extended_graph import CityNode, StopNode, RouteNode, ExtendedGraph, ExtendedEdgesType, ExtendedEdge, ExtendedNode
from .hyper_path import Hyperpath
from .assignment import Assignment

__all__ = ['Assignment', 'ExtendedGraph', 'ExtendedEdgesType', 'Hyperpath', 'CityNode', 'StopNode', 'RouteNode',
           'ExtendedEdge', 'ExtendedNode']



from __future__ import annotations

import math
from collections import defaultdict
from enum import Enum
from typing import List

import networkx as nx
import pandas as pd
from matplotlib import pyplot as plt

from sidermit.exceptions import *


class GraphContentFormat(Enum):
    """
    Supported format for reading files in the construction of the city graph.
    Pajek reference: https://gephi.org/users/supported-graph-formats/pajek-net-format/
    Defined format: [node_id] [node name] [x] [y] [type] [zone] [width]
    """
    PAJEK = 1


class Node:

    def __init__(self, node_id: int, x: float, y: float, radius: float, angle: float, width: float, zone_id: int,
                 name: str):
        """
        Graph node
        :param node_id: node id, must be unique in the graph
        :param x: horizontal position in meters of the node in the Cartesian plane
        :param y: vertical position in meters of the node in the Cartesian plane
        :param radius: node radius in meters from point 0,0 of the Cartesian plane
        :param angle: angle in degrees with range [0° - 360°] measured from x+ axis counterclockwise
        :param width: node width in meters, takes part in the demand assignment considering lateral access time
        :param zone_id: zone id that the node belongs
        :param name: node name
        """
        if name is None:
            raise NameIsNotDefinedException("must define a name to node")
        if node_id is None:
            raise NodeIdIsNotValidException("zone_id is not valid")
        if zone_id is None:
            raise ZoneIdIsNotValidException("zone_id is not valid")
        if radius < 0:
            raise NodeRadiusIsNotValidException("radius must be positive")
        if angle < 0 or angle > 360:
            raise NodeAngleIsNotValidException("angle must belong into range [0°-360°]")
        if width < 0:
            raise NodeWidthIsNotValidException("width must be >= 0")
        self.id = node_id
        self.x = x
        self.y = y
        self.radius = radius
        self.angle = angle
        self.width = width
        self.name = name
        self.zone_id = zone_id


class CBD(Node):

    def __init__(self, node_id: int, x: float, y: float, radius: float, angle: float, width: float, zone_id: int,
                 name: str):
        """
        CBD node
        :param node_id: node id, must be unique in the graph
        :param x: horizontal position in meters of the node in the Cartesian plane
        :param y: vertical position in meters of the node in the Cartesian plane
        :param radius: node radius in meters from point 0,0 of the Cartesian plane
        :param angle: angle in degrees with range [0° - 360°] measured from x+ axis counterclockwise
        :param width: node width in meters, takes part in the demand assignment considering lateral access time
        :param zone_id: zone id that the node belongs. CBD must belong to zone with id 0
        :param name: node name
        """
        if zone_id != 0:
            raise ZoneIdIsNotValidException("CBD node_id must be equal to 0")
        Node.__init__(self, node_id, x, y, radius, angle, width, zone_id, name)


class Periphery(Node):

    def __init__(self, node_id: int, x: float, y: float, radius: float, angle: float, width: float, zone_id: int,
                 name: str):
        """
        Periphery node
        :param node_id: node id, must be unique in the graph
        :param x: horizontal position in meters of the node in the Cartesian plane
        :param y: vertical position in meters of the node in the Cartesian plane
        :param radius: node radius in meters from point 0,0 of the Cartesian plane
        :param angle: angle in degrees with range [0° - 360°] measured from x+ axis counterclockwise
        :param width: node width in meters, takes part in the demand assignment considering lateral access time
        :param zone_id: zone id that the node belongs
        :param name: node name
        """
        Node.__init__(self, node_id, x, y, radius, angle, width, zone_id, name)


class Subcenter(Node):

    def __init__(self, node_id: int, x: float, y: float, radius: float, angle: float, width: float, zone_id: int,
                 name: str):
        """
        Subcenter node
        :param node_id: node id, must be unique in the graph
        :param x: horizontal position in meters of the node in the Cartesian plane
        :param y: vertical position in meters of the node in the Cartesian plane
        :param radius: node radius in meters from point 0,0 of the Cartesian plane
        :param angle: angle in degrees with range [0° - 360°] measured from x+ axis counterclockwise
        :param width: node width in meters, takes part in the demand assignment considering lateral access time
        :param zone_id: zone id that the node belongs
        :param name: node name
        """
        Node.__init__(self, node_id, x, y, radius, angle, width, zone_id, name)


class Zone:

    def __init__(self, zone_id: int, node_periphery: Periphery, node_subcenter: Subcenter):
        """
        Area of the city, they have two nodes, one of them corresponding to the subcenter and another to the periphery.
        :param zone_id: zone id must be unique in the city graph and belong to an ordered list of id [1, ..., n]
        :param node_periphery:
        :param node_subcenter:
        """

        if zone_id is None:
            raise ZoneIdIsNotValidException("zone_id is not valid")
        if zone_id <= 0:
            raise ZoneIdIsNotValidException("zone_id number must be >=1")
        if not isinstance(node_periphery, Periphery):
            raise PeripheryNodeTypeIsNotValidException('node must be periphery')

        if not isinstance(node_subcenter, Subcenter):
            raise SubcenterNodeTypeIsNotValidException('node must be subcenter')

        if node_periphery.zone_id != zone_id or node_subcenter.zone_id != zone_id:
            raise ZoneIdIsNotValidException("node_periphery and node_subcenter must be equal to zone_id")

        self.id = zone_id
        self.periphery = node_periphery
        self.subcenter = node_subcenter


class Edge:

    def __init__(self, edge_id: int, node1, node2):
        """
        Graph edge. Allowed edges are: [Periphery - Subcenter], [Subcenter - Periphery], [Subcenter - Subcenter],
        [Subcenter - CBD], [CBD - Subcenter]
        :param edge_id: edge id must be unique in the city graph
        :param node1: edge start node
        :param node2: edge end node
        """
        if edge_id is None:
            raise EdgeIdIsNotValidException("edge_id is not valid")
        if isinstance(node1, Periphery) and isinstance(node2, Periphery):
            raise EdgeIsNotAvailableException("edge between periphery nodes is not available")
        if isinstance(node1, Periphery) and isinstance(node2, CBD):
            raise EdgeIsNotAvailableException("edge between periphery and CBD is not available")
        if isinstance(node1, CBD) and isinstance(node2, Periphery):
            raise EdgeIsNotAvailableException("edge between periphery and CBD is not available")
        self.id = edge_id
        self.node1 = node1
        self.node2 = node2


defaultdict2_float = defaultdict(lambda: defaultdict(float))
coor = (float, float)
vector = List[float]
parameters = (int, float, float, float, float, int, vector, vector, vector)
zones = List[Zone]
nodes = List[Node]
edges = List[Edge]


class Graph:

    def __init__(self):
        """
        Graph of the city. Defined by n zones, n periphery nodes, n subcenter nodes, 1 CBD node, and 6 * n edges.
        The edges are defined by construction between the periphery and subcenter of the same zone, the subcenter and
        the CBD of all zones and between adjacent sub-centers of zones
        """

        self.__nodes_id = []
        self.__edges_id = []
        self.__cbd_exist = False

        self.__zones = []
        self.__nodes = []
        self.__edges = []

        self.__n = None
        self.__l = None
        self.__g = None
        self.__p = None
        self.__etha = None
        self.__etha_zone = None
        self.__angles = None
        self.__Gi = None
        self.__Hi = None

    def edge_exist(self, id_i, id_j) -> bool:
        """
        to check if edge with origin node id = id_i and destination node id = id_j exist
        :param id_i:
        :param id_j:
        :return: True if edge exist, False if not.
        """
        for edge in self.__edges:
            if edge.node1.id == id_i and edge.node2.id == id_j:
                return True
        return False

    def export_graph(self, format: GraphContentFormat = GraphContentFormat.PAJEK) -> str:
        """
        to get a string with graph nodes information in a specific format file
        :param format: GraphContentFormat
        :return: string with graph nodes information in a specific format file
        """

        line = ""
        if format == GraphContentFormat.PAJEK:
            line = "*vertices {}{}".format(len(self.__nodes), "\n")

            for node_obj in self.__nodes:
                node_type = None
                if isinstance(node_obj, CBD):
                    node_type = 'CBD'
                elif isinstance(node_obj, Periphery):
                    node_type = 'P'
                elif isinstance(node_obj, Subcenter):
                    node_type = 'SC'

                line += "{0} {1} {2} {3} {4} {5} {6}\n".format(node_obj.id, node_obj.name, node_obj.x, node_obj.y,
                                                               node_type, node_obj.zone_id,
                                                               node_obj.width)
        return line

    def graph_to_pajek_file(self, file_path, format: GraphContentFormat = GraphContentFormat.PAJEK):
        """
        write city graph data to file using Pajek format
        :param format:
        :param file_path: file path
        :return: None
        """

        file = open(file_path, 'w')
        string_file = self.export_graph(format)
        file.write(string_file)

        file.close()

    @staticmethod
    def parameters_validator(n: int, l: float, g: float, p: float, etha=None, etha_zone=None, angles=None, Gi=None,
                             Hi=None) -> bool:
        """
        to validate symmetric and asymmetric graph construction parameters.
        :param n: zone numbers.
        :param l: average distance in meter of subcenters to (0,0) of the Cartesian plane.
        :param g: (>=0) to represent average distance of peripheries (= l + g * l).
        :param p: node width in meters, takes part in the demand assignment considering lateral access time.
        :param etha: [0-1] CBD eccentricity ratio.
        :param etha_zone: CBD eccentricity direction zone
        :param angles: list of construction angles of the zones. You must specify an angle for each zone.
        Angle in degrees with range [0° - 360°] measured from x+ axis counterclockwise
        :param Gi: asymmetry parameter of the distance from the subcenters to (0,0) of the Cartesian plane.
        List with asymmetry ratio for each zone with values >=0
        :param Hi: asymmetry parameter of the distance from the peripheries to (0,0) of the Cartesian plane.
        List with asymmetry ratio for each zone with values >=0
        :return: True if parameters are valid. False if not.
        """
        if n < 0 or not isinstance(n, int):
            raise NIsNotValidNumberException('n cannot be a negative number')
        if l < 0:
            raise LIsNotValidNumberException('L cannot be a negative number')
        if g < 0:
            raise GIsNotValidNumberException('G cannot be a negative number')
        if p < 0:
            raise PIsNotValidNumberException('n cannot be a negative number')
        if etha is None and etha_zone is not None:
            raise EthaValueRequiredException("must give value for etha")
        if etha is not None and etha_zone is None:
            raise EthaZoneValueRequiredException("must give value for etha zone")
        if etha is not None and etha_zone is not None:
            if etha < 0 or etha > 1:
                raise EthaValueIsNotValidException("etha value is not valid. Try with value belong in [0-1]")
            elif etha_zone <= 0 or etha_zone > n or not isinstance(etha_zone, int):
                raise EthaZoneValueIsNotValidException("etha zone is not valid. Try with value belong in [1,...,n]")
        if angles is not None:
            if len(angles) != n:
                raise AngleListLengthIsNotValidException("must give angle value for all zones")
            for angle in angles:
                if angle < 0 or angle > 360:
                    raise AngleValueIsNotValidException("angle must belong in range [0°-360°]")
        if Gi is not None:
            if len(Gi) != n:
                raise GiListLengthIsNotValidException("must give Gi value for all zones")
            for gi in Gi:
                if gi < 0:
                    raise GiValueIsNotValidException("Gi must be >= 0")
        if Hi is not None:
            if len(Hi) != n:
                raise HiListLengthIsNotValidException("must give Hi value for all zones")
            for hi in Hi:
                if hi < 0:
                    raise HiValueIsNotValidException("Hi must be >= 0")
        return True

    @staticmethod
    def build_from_parameters(n: int = 2, l: float = 10, g: float = 0.5, p: float = 2, etha=None, etha_zone=None,
                              angles=None, Gi=None, Hi=None) -> Graph:
        """
        to build a city graph with parameters information
        :param n: zone numbers.
        :param l: average distance in meter of subcenters to (0,0) of the Cartesian plane.
        :param g: (>=0) to represent average distance of peripheries (= l + g * l).
        :param p: node width in meters, takes part in the demand assignment considering lateral access time.
        :param etha: [0-1] CBD eccentricity ratio.
        :param etha_zone: CBD eccentricity direction zone
        :param angles: list of construction angles of the zones. You must specify an angle for each zone.
        Angle in degrees with range [0° - 360°] measured from x+ axis counterclockwise
        :param Gi: asymmetry parameter of the distance from the subcenters to (0,0) of the Cartesian plane.
        List with asymmetry ratio for each zone with values >=0
        :param Hi: asymmetry parameter of the distance from the peripheries to (0,0) of the Cartesian plane.
        List with asymmetry ratio for each zone with values >=0
        :return: Graph
        """

        graph_obj = Graph()

        # if parameters are valid
        if graph_obj.parameters_validator(n, l, g, p, etha, etha_zone, angles, Gi, Hi):
            # ADD CBD
            # ADD Zones
            # # for each zones add Nodes periphery and subcenter
            # ADD all edges
            # add CBD
            cbd_node_obj = CBD(0, 0, 0, 0, 0, p, 0, "CBD")
            graph_obj.__add_cbd(cbd_node_obj)

            # Add zones
            for zone in range(n):
                periphery_node_id = 2 * zone + 1
                subcenter_node_id = 2 * zone + 2
                periphery_radius = l + g * l
                subcenter_radius = l
                periphery_angle = 360 / n * zone
                subcenter_angle = 360 / n * zone
                x_p, y_p = graph_obj.get_xy(periphery_radius, periphery_angle)
                x_sc, y_sc = graph_obj.get_xy(subcenter_radius, subcenter_angle)

                periphery_node_name = "P_{}".format(zone + 1)
                periphery_node_obj = Periphery(periphery_node_id, x_p, y_p, periphery_radius, periphery_angle, p,
                                               zone + 1, periphery_node_name)
                subcenter_node_name = "SC_{}".format(zone + 1)
                subcenter_node_obj = Subcenter(subcenter_node_id, x_sc, y_sc, subcenter_radius, subcenter_angle, p,
                                               zone + 1, subcenter_node_name)

                # add periphery and subcenter nodes to graph obj
                graph_obj.__add_zone(periphery_node_obj, subcenter_node_obj, zone + 1)

            # add asymmetry by angles
            if angles is not None:
                graph_obj.__angles_asymmetry(angles)
            # add asymmetry by Gi
            if Gi is not None:
                graph_obj.__Gi_asymmetry(Gi)
            # add asymmetry by Hi
            if Hi is not None:
                graph_obj.__Hi_asymmetry(Hi)
            # add asymmetry by etha
            if etha is not None and etha_zone is not None:
                graph_obj.__etha_asymmetry(etha, etha_zone)

        graph_obj.__n = n
        graph_obj.__l = l
        graph_obj.__g = g
        graph_obj.__p = p
        graph_obj.__etha = etha
        graph_obj.__etha_zone = etha_zone
        graph_obj.__angles = angles
        graph_obj.__Gi = Gi
        graph_obj.__Hi = Hi

        return graph_obj

    @staticmethod
    def __pajekfile_to_dataframe(data) -> pd.DataFrame:
        """
        convert string pajek to dataframe
        :param data: string data in  pajek format
        :return: Dataframe panda
        """

        # read pajek file as a Dataframe
        df_file = pd.DataFrame()
        col_id = []
        col_name = []
        col_x = []
        col_y = []
        col_type = []
        col_zone = []
        col_width = []

        f_obj = data.split("\n")

        # print(f_obj)

        n_nodes = 0
        for line in f_obj:
            if line.lower().startswith("*vertices"):
                _, n_nodes = line.split()
                n_nodes = int(n_nodes)
                continue
            if n_nodes > 0:
                if len(line.split()) == 7:
                    node_id, name, x, y, node_type, zone, width = line.split()
                    if node_type != "CBD" and node_type != "SC" and node_type != "P":
                        raise NodeTypeIsNotValidException("Node type is not valid. Try with CBD, SC or P")

                    col_name.append(name)
                    col_x.append(x)
                    col_y.append(y)
                    col_type.append(node_type)
                    col_width.append(width)

                    try:
                        col_id.append(int(node_id))
                    except ValueError:
                        raise NodeIdIsNotAnInteger("node id must be a integer")
                    try:
                        col_zone.append(int(zone))
                    except ValueError:
                        raise ZoneIdIsNotAnInteger("zone id must be a integer")

                    n_nodes = n_nodes - 1
                else:
                    raise PajekFormatIsNotValidException("each node line must provide information about [id] ["
                                                         "name] [x] [y] [type] [zone] [width]")

        df_file["node_id"] = col_id
        df_file["name"] = col_name
        df_file["x"] = col_x
        df_file["y"] = col_y
        df_file["type"] = col_type
        df_file["zone"] = col_zone
        df_file["width"] = col_width

        return df_file

    def get_edges_distance(self) -> defaultdict2_float:
        """
        to get a dictionary with the edge distance of a nodes pair
        :return: dic[nodei_id][nodej_id] = float [km]; defaultdict(lambda: defaultdict(float))
        """

        edge_distance = defaultdict(lambda: defaultdict(float))

        for edge in self.__edges:
            nodei = edge.node1
            nodej = edge.node2

            xi = nodei.x
            yi = nodei.y
            xj = nodej.x
            yj = nodej.y

            edge_distance[nodei.id][nodej.id] = ((xi - xj) ** 2 + (yi - yj) ** 2) ** 0.5

        return edge_distance

    @staticmethod
    def get_angle(x: float, y: float) -> float:
        """
        to get angle of a vector with coor (0,0,x,y)  with respect to x+
        :param x:
        :param y:
        :return: angle in degrees with range[0° - 360°] measured from x+ axis counterclockwise
        """
        a = 1
        b = 0
        c = x
        d = y

        if x == 0 and y == 0:
            return 0

        dot_product = a * c + b * d
        # for three dimensional simply add dotProduct = a*c + b*d  + e*f
        mod_of_vector1 = math.sqrt(a * a + b * b) * math.sqrt(c * c + d * d)
        # for three dimensional simply add modOfVector = math.sqrt( a*a + b*b + e*e)*math.sqrt(c*c + d*d +f*f)
        angle = dot_product / mod_of_vector1
        # print("Cosθ =",angle)
        angle_in_degree = math.degrees(math.acos(angle))
        if y < 0:
            if x >= 0:
                angle_in_degree = 360 - angle_in_degree
            if x < 0:
                angle_in_degree = (180 - angle_in_degree) + 180
        return angle_in_degree

    @staticmethod
    def get_xy(radius, angle) -> coor:
        """
        to get x, y with radius and angle parameters
        :param radius:
        :param angle:
        :return: x,y in the Cartesian plane
        """
        x, y = radius * math.cos(math.radians(angle)), radius * math.sin(math.radians(angle))
        return x, y

    @staticmethod
    def build_from_content(data, content_format: GraphContentFormat = GraphContentFormat.PAJEK) -> Graph:
        """
        :param data:
        :param content_format:
        :return: Graph
        """
        graph_obj = Graph()

        if content_format == GraphContentFormat.PAJEK:

            df_file = graph_obj.__pajekfile_to_dataframe(data)

            df_file = df_file.sort_values(["zone", "type"], ascending=[True, True])

            lines_accepted = list(range(1, 10001, 2))

            if not len(df_file["node_id"]) in lines_accepted:
                # print(len(df_file["node_id"]))
                raise LineNumberInFileIsNotValidException("The number of lines in the file must be 2n+1 or "
                                                          "file is very big (until 5000 zones accepted)")

            n = int((len(df_file["node_id"]) - 1) / 2)

            cbd = []
            periphery = []
            subcenter = []

            width = 0

            # build nodes list
            for i in range(len(df_file["node_id"])):
                node_id = df_file.iloc[i]["node_id"]
                name = df_file.iloc[i]["name"]
                x = float(df_file.iloc[i]["x"])
                y = float(df_file.iloc[i]["y"])
                node_type = df_file.iloc[i]["type"]
                zone = int(df_file.iloc[i]["zone"])
                width = float(df_file.iloc[i]["width"])

                radius = math.sqrt(x ** 2 + y ** 2)
                angle = graph_obj.get_angle(x, y)

                if node_type == "CBD":
                    cbd.append(CBD(node_id, x, y, radius, angle, width, zone, name))
                if node_type == "P":
                    periphery.append(Periphery(node_id, x, y, radius, angle, width, zone, name))
                if node_type == "SC":
                    subcenter.append(Subcenter(node_id, x, y, radius, angle, width, zone, name))

            r_sc = []
            r_p = []
            ang_sc = []
            r_cbd = 0
            ang_cbd = 0

            if len(cbd) == 1:
                r_cbd = cbd[0].radius
                ang_cbd = cbd[0].angle

            etha = None
            etha_zone = None

            # verification of existence and uniqueness of SC and P by zone
            # save information to build parameters
            for zone in range(n):
                zone_id = zone + 1
                n_p = 0
                n_sc = 0
                node_periphery = None
                node_subcenter = None

                for p in periphery:
                    if p.zone_id == zone_id:
                        n_p = n_p + 1
                        node_periphery = p
                for sc in subcenter:
                    if sc.zone_id == zone_id:
                        n_sc = n_sc + 1
                        node_subcenter = sc
                if n_p != 1 or n_sc != 1:
                    raise PeripherySubcenterNumberForZoneException("try to verify that each zone [1, ..., n] has "
                                                                   "one sc and p")

                radius_p = node_periphery.radius
                radius_sc = node_subcenter.radius
                angle_sc = node_subcenter.angle

                r_sc.append(radius_sc)
                r_p.append(radius_p)
                ang_sc.append(angle_sc)

                # Update etha and etha_zone
                if ang_cbd - 0.001 <= angle_sc <= ang_cbd + 0.001:
                    etha_zone = zone_id
                    etha = r_cbd / radius_sc

            L = 0
            g = 0
            if len(r_sc) != 0 and len(r_p) != 0:
                L = sum(r_sc) / len(r_sc)
                g = (sum(r_p) / len(r_p) - L) / L

            Gi = []
            Hi = []
            for i in range(n):
                Gi.append(r_sc[i] / L)
                Hi.append(r_p[i] / (L + g * L))

            angles = ang_sc

            # if parameters are valid
            if graph_obj.parameters_validator(n, L, g, width, etha, etha_zone, angles, Gi, Hi):

                for node in cbd:
                    graph_obj.__add_cbd(node)

                for i in range(n):
                    node_periphery = None
                    node_subcenter = None
                    for p in periphery:
                        if p.zone_id == i + 1:
                            node_periphery = p
                            break
                    for sc in subcenter:
                        if sc.zone_id == i + 1:
                            node_subcenter = sc
                            break
                    graph_obj.__add_zone(node_periphery, node_subcenter)

        else:
            raise FileFormatIsNotValidException("File don't have a valid format. Try with Pajek format")

        graph_obj.__n = n
        graph_obj.__l = L
        graph_obj.__g = g
        graph_obj.__p = width
        graph_obj.__etha = etha
        graph_obj.__etha_zone = etha_zone
        graph_obj.__angles = angles
        graph_obj.__Gi = Gi
        graph_obj.__Hi = Hi

        return graph_obj

    @staticmethod
    def build_from_file(file_path, file_format=GraphContentFormat.PAJEK) -> Graph:
        """
        to build a city graph with pajek file information
        :param file_path:
        :param file_format: default and only supported value is PAJEK
        :return: Graph instance
        """

        graph_obj = Graph()

        with open(file_path, mode='r', encoding='utf-8') as f_obj:
            data = f_obj.read()
            # print(data)

            graph_obj = graph_obj.build_from_content(data, file_format)

        return graph_obj

    def __add_cbd(self, cbd_node):
        """
        to add cbd node in the graph
        :param cbd_node:
        :return:
        """
        self.__cbd_exist = True
        self.__nodes_id.append(cbd_node.id)
        self.__nodes.append(cbd_node)

    def __add_zone(self, node_periphery: Periphery, node_subcenter: Subcenter, zone_id=None):
        """
        to add zone in the graph
        :param node_periphery:
        :param node_subcenter:
        :param zone_id:
        :return:
        """
        # zones can only be added if CBD is included
        if len(self.__nodes) == 0:
            raise CBDDoesNotExistException("Need add CBD node previously")
        # always add last zone
        if zone_id is None:
            zone_id = len(self.__zones) + 1
        # not need to check because number of periphery and subcenter control exceptions
        # if zone_id != len(self.__zones) + 1:
        #    raise AddPreviousZonesExceptions("need to specify zone {} previously".format(len(self.__zones) + 1))
        self.__zones.append(Zone(zone_id, node_periphery, node_subcenter))
        self.__add_nodes(node_periphery, node_subcenter)

        # must build all the edges
        self.__build_edges()

    def __add_nodes(self, node_periphery: Periphery, node_subcenter: Subcenter):
        """
        to add periphery and subcenter belonging to a zone
        :param node_periphery:
        :param node_subcenter:
        :return:
        """

        # to verified if id is not duplicated
        if node_periphery.id in self.__nodes_id or node_subcenter.id in self.__nodes_id:
            raise NodeIdDuplicatedException('id_node is duplicated')

        self.__nodes.append(node_periphery)
        self.__nodes.append(node_subcenter)
        self.__nodes_id.append(node_periphery.id)
        self.__nodes_id.append(node_subcenter.id)

    def __build_edges(self):
        """
        to build all edges defined by construction in the graph. The edges are defined by construction between the
        periphery and subcenter of the same zone, the subcenter and the CBD of all zones and between adjacent
        sub-centers of zones
        :return:
        """
        self.__edges = []
        self.__edges_id = []
        if len(self.__zones) == 1:
            # p <-> sc
            # sc <-> cbd
            p = self.__zones[0].periphery
            sc = self.__zones[0].subcenter
            cbd = self.__nodes[0]
            self.__add_edge(Edge(len(self.__edges) + 1, p, sc))
            self.__add_edge(Edge(len(self.__edges) + 1, sc, p))
            self.__add_edge(Edge(len(self.__edges) + 1, sc, cbd))
            self.__add_edge(Edge(len(self.__edges) + 1, cbd, sc))
        if len(self.__zones) == 2:
            # p <-> sc
            # sc <-> cbd
            p1 = self.__zones[0].periphery
            sc1 = self.__zones[0].subcenter
            p2 = self.__zones[1].periphery
            sc2 = self.__zones[1].subcenter
            cbd = self.__nodes[0]
            self.__add_edge(Edge(len(self.__edges) + 1, p1, sc1))
            self.__add_edge(Edge(len(self.__edges) + 1, sc1, p1))
            self.__add_edge(Edge(len(self.__edges) + 1, sc1, cbd))
            self.__add_edge(Edge(len(self.__edges) + 1, cbd, sc1))
            self.__add_edge(Edge(len(self.__edges) + 1, sc1, sc2))
            self.__add_edge(Edge(len(self.__edges) + 1, sc2, sc1))
            self.__add_edge(Edge(len(self.__edges) + 1, p2, sc2))
            self.__add_edge(Edge(len(self.__edges) + 1, sc2, p2))
            self.__add_edge(Edge(len(self.__edges) + 1, sc2, cbd))
            self.__add_edge(Edge(len(self.__edges) + 1, cbd, sc2))

        if len(self.__zones) > 2:
            for i in range(len(self.__zones)):
                p = self.__zones[i].periphery
                sc = self.__zones[i].subcenter
                cbd = self.__nodes[0]
                j = i + 1
                if i == len(self.__zones) - 1:
                    j = 0
                sc2 = self.__zones[j].subcenter
                # p <-> sc
                # sc <-> cbd
                # sc <-> sc2
                self.__add_edge(Edge(len(self.__edges) + 1, p, sc))
                self.__add_edge(Edge(len(self.__edges) + 1, sc, p))
                self.__add_edge(Edge(len(self.__edges) + 1, sc, cbd))
                self.__add_edge(Edge(len(self.__edges) + 1, cbd, sc))
                self.__add_edge(Edge(len(self.__edges) + 1, sc, sc2))
                self.__add_edge(Edge(len(self.__edges) + 1, sc2, sc))

    def __add_edge(self, edge: Edge):
        """
        to add a edge in the graph
        :param edge:
        :return:
        """
        self.__edges.append(edge)
        self.__edges_id.append(edge.id)

    def __etha_asymmetry(self, etha, etha_zone):
        """
        modification of the graph incorporating eccentricity of the CBD
        :param etha: [0-1] CBD eccentricity ratio.
        :param etha_zone: CBD eccentricity direction zone
        :return:
        """

        zone = self.__zones[etha_zone - 1]
        sc = zone.subcenter
        r_cbd = etha * sc.radius
        ang = sc.angle

        x_cbd, y_cbd = self.get_xy(r_cbd, ang)

        for node in range(len(self.__nodes)):
            if isinstance(self.__nodes[node], CBD):
                self.__nodes[node].x = x_cbd
                self.__nodes[node].y = y_cbd
                self.__nodes[node].radius = r_cbd
                self.__nodes[node].angle = ang

        # must build all the edges
        self.__build_edges()

    def __Hi_asymmetry(self, Hi):
        """
        graph modification incorporating asymmetry of the peripheries
        :param Hi: asymmetry parameter of the distance from the peripheries to (0,0) of the Cartesian plane.
        List with asymmetry ratio for each zone with values >=0
        :return:
        """

        for i in range(len(self.__zones)):
            hi = Hi[i]
            p = self.__zones[i].periphery

            r_p = hi * p.radius
            ang = p.angle
            x_p, y_p = self.get_xy(r_p, ang)

            # change values in nodes list
            for node in range(len(self.__nodes)):

                if self.__nodes[node] == p:
                    self.__nodes[node].x = x_p
                    self.__nodes[node].y = y_p
                    self.__nodes[node].radius = r_p

            # change values in zones nodes information
            self.__zones[i].periphery.x = x_p
            self.__zones[i].periphery.y = y_p
            self.__zones[i].periphery.radius = r_p

        # must build all the edges
        self.__build_edges()

    def __Gi_asymmetry(self, Gi):

        """
        graph modification incorporating asymmetry of the subcenters
        :param Gi: asymmetry parameter of the distance from the subcenters to (0,0) of the Cartesian plane.
        List with asymmetry ratio for each zone with values >=0
        :return:
        """

        for i in range(len(self.__zones)):
            gi = Gi[i]
            sc = self.__zones[i].subcenter

            r_sc = gi * sc.radius
            ang = sc.angle
            x_sc, y_sc = self.get_xy(r_sc, ang)

            # change values in nodes list
            for node in range(len(self.__nodes)):

                if self.__nodes[node] == sc:
                    self.__nodes[node].x = x_sc
                    self.__nodes[node].y = y_sc
                    self.__nodes[node].radius = r_sc

            # change values in zones nodes information
            self.__zones[i].subcenter.x = x_sc
            self.__zones[i].subcenter.y = y_sc
            self.__zones[i].subcenter.radius = r_sc

        # must build all the edges
        self.__build_edges()

    def __angles_asymmetry(self, angles):
        """
        graph modification incorporating angular asymmetry of the zones
        :param angles: list of construction angles of the zones. You must specify an angle for each zone.
        Angle in degrees with range [0° - 360°] measured from x+ axis counterclockwise
        :return:
        """

        for i in range(len(self.__zones)):
            ang = angles[i]

            p = self.__zones[i].periphery
            sc = self.__zones[i].subcenter

            r_p = p.radius
            r_sc = sc.radius
            x_p, y_p = self.get_xy(r_p, ang)
            x_sc, y_sc = self.get_xy(r_sc, ang)

            # change values in nodes list
            for node in range(len(self.__nodes)):
                if self.__nodes[node] == p:
                    self.__nodes[node].x = x_p
                    self.__nodes[node].y = y_p
                    self.__nodes[node].angle = ang
                if self.__nodes[node] == sc:
                    self.__nodes[node].x = x_sc
                    self.__nodes[node].y = y_sc
                    self.__nodes[node].angle = ang

            # change values in zones nodes information
            self.__zones[i].periphery.x = x_p
            self.__zones[i].periphery.y = y_p
            self.__zones[i].periphery.angle = ang

            self.__zones[i].subcenter.x = x_sc
            self.__zones[i].subcenter.y = y_sc
            self.__zones[i].subcenter.angle = ang

        # must build all the edges
        self.__build_edges()

    def get_cbd(self) -> CBD:
        """
        to get CBD node defined in the graph
        :return: CBD node
        """
        cbd = None
        for node in self.__nodes:
            if isinstance(node, CBD):
                cbd = node
                break
        return cbd

    def get_parameters(self) -> parameters:
        """
        to get tuple with symmetric and asymmetric parameters n, L, g, P, etha, etha_zone, angles, Gi, Hi
        :return: n, L, g, P, etha, etha_zone, angles, Gi, Hi
        """
        return (self.__n, self.__l, self.__g, self.__p, self.__etha,
                self.__etha_zone, self.__angles, self.__Gi, self.__Hi)

    def get_n(self) -> int:
        """
        to get parameters of numbers of zones
        :return: n
        """
        return self.__n

    def get_zones(self) -> zones:
        """
        to get all the zones built
        :return: list of zones
        """
        return self.__zones

    def get_nodes(self) -> nodes:
        """
        to get all the nodes built
        :return: list of nodes
        """
        return self.__nodes

    def get_edges(self) -> edges:
        """
        to get all the edges built
        :return: list of edges
        """
        return self.__edges

    def plot(self, file_path):
        """
        to plot graph defined
        :param file_path:
        :return:
        """

        # edges information and positions
        edges = []
        position = defaultdict(list)
        for edge in self.__edges:
            edges.append((edge.node1.id, edge.node2.id))
            if not position.get(edge.node1.id):
                position[edge.node1.id].append(edge.node1.x)
                position[edge.node1.id].append(edge.node1.y)
            if not position.get(edge.node2.id):
                position[edge.node2.id].append(edge.node2.x)
                position[edge.node2.id].append(edge.node2.y)

        # nodes information and positions
        info_cbd = []
        info_sc = []
        info_p = []

        for node in self.__nodes:
            if isinstance(node, CBD):
                info_cbd.append(node)
            if isinstance(node, Periphery):
                info_p.append(node)
            if isinstance(node, Subcenter):
                info_sc.append(node)
        id_cbd = []
        id_sc = []
        id_p = []
        x_cbd = []
        y_cbd = []
        x_sc = []
        y_sc = []
        x_p = []
        y_p = []

        for cbd in info_cbd:
            x_cbd.append(cbd.x)
            y_cbd.append(cbd.y)
            id_cbd.append(cbd.id)

        for sc in info_sc:
            x_sc.append(sc.x)
            y_sc.append(sc.y)
            id_sc.append(sc.id)
        for p in info_p:
            x_p.append(p.x)
            y_p.append(p.y)
            id_p.append(p.id)

        G = nx.DiGraph()
        G.add_edges_from(edges)

        # separate calls to draw labels, nodes and edges
        nx.draw_networkx_nodes(G, position, cmap=plt.get_cmap('Set2'), nodelist=id_p, node_color='red', node_size=300)
        nx.draw_networkx_nodes(G, position, cmap=plt.get_cmap('Set2'), nodelist=id_sc, node_color='blue', node_size=300)
        nx.draw_networkx_nodes(G, position, cmap=plt.get_cmap('Set2'), nodelist=id_cbd, node_color='purple',
                               node_size=300)
        nx.draw_networkx_labels(G, position)
        nx.draw_networkx_edges(G, position, edgelist=edges, edge_color='orange', arrows=True)

        plt.title("City graph")
        plt.xlabel("X")
        plt.ylabel("Y")
        plt.gca().set_aspect('equal')
        plt.savefig(file_path)



import unittest

from sidermit import exceptions
from sidermit.publictransportsystem import TransportMode, TransportModeManager


class TransportModeTest(unittest.TestCase):

    def test_is_valid_to_assignment_step(self):
        """
        test is_valid method
        :return:
        """

        m = TransportModeManager()

        self.assertTrue(m.is_valid_to_assignment_step())

        m.remove_mode("bus")

        self.assertTrue(m.is_valid_to_assignment_step())

        m.remove_mode("metro")

        self.assertTrue(not m.is_valid_to_assignment_step())

    def test_raises_mode_exceptions(self):
        """
        to test mode class exceptions
        :return:
        """
        name = "bus"
        bya = 1
        co = 1
        c1 = 1
        c2 = 1
        v = 20
        t = 0.2
        fmax = 100
        kmax = 50
        theta = 0.5
        tat = 1
        d = 1
        fini = 28

        with self.assertRaises(exceptions.NameIsNotValidExceptions):
            TransportMode(None, bya, co, c1, c2, v, t, fmax, kmax, theta, tat,
                          d, fini)
        with self.assertRaises(exceptions.ByaIsNotValidExceptions):
            TransportMode(name, -1, co, c1, c2, v, t, fmax, kmax, theta, tat,
                          d, fini)
        with self.assertRaises(exceptions.CoIsNotValidExceptions):
            TransportMode(name, bya, -2, c1, c2, v, t, fmax, kmax, theta, tat,
                          d, fini)
        with self.assertRaises(exceptions.C1IsNotValidExceptions):
            TransportMode(name, bya, co, None, c2, v, t, fmax, kmax, theta,
                          tat, d, fini)
        with self.assertRaises(exceptions.C2IsNotValidExceptions):
            TransportMode(name, bya, co, c1, -3, v, t, fmax, kmax, theta, tat,
                          d, fini)
        with self.assertRaises(exceptions.VIsNotValidExceptions):
            TransportMode(name, bya, co, c1, c2, None, t, fmax, kmax, theta,
                          tat, d, fini)
        with self.assertRaises(exceptions.TIsNotValidExceptions):
            TransportMode(name, bya, co, c1, c2, v, -3, fmax, kmax, theta, tat,
                          d, fini)
        with self.assertRaises(exceptions.FmaxIsNotValidExceptions):
            TransportMode(name, bya, co, c1, c2, v, t, -200, kmax, theta, tat,
                          d, fini)
        with self.assertRaises(exceptions.KmaxIsNotValidExceptions):
            TransportMode(name, bya, co, c1, c2, v, t, fmax, -100, theta, tat,
                          d, fini)
        with self.assertRaises(exceptions.ThetaIsNotValidExceptions):
            TransportMode(name, bya, co, c1, c2, v, t, fmax, kmax, 2, tat, d, fini)
        with self.assertRaises(exceptions.TatIsNotValidExceptions):
            TransportMode(name, bya, co, c1, c2, v, t, fmax, kmax, theta, None,
                          d, fini)
        with self.assertRaises(exceptions.DIsNotValidExceptions):
            TransportMode(name, bya, co, c1, c2, v, t, fmax,
                          kmax, theta, tat, -1, fini)

        with self.assertRaises(exceptions.FiniIsNotValidException):
            TransportMode(name, bya, co, c1, c2, v, t, fmax,
                          kmax, theta, tat, d, -1)

    def test_get(self):
        """
        to test get methods
        :return:
        """
        m = TransportModeManager()

        self.assertEqual(len(m.get_modes()), 2)
        self.assertEqual(len(m.get_modes_names()), 2)
        self.assertTrue(
            isinstance(m.get_mode("bus"), TransportMode))

        with self.assertRaises(exceptions.ModeNotFoundExceptions):
            m.get_mode("train")

    def test_add_mode(self):
        """
        to test add_mode method
        :return:
        """
        m = TransportModeManager()

        name = "train"
        bya = 1
        co = 1
        c1 = 1
        c2 = 1
        v = 20
        t = 0.2
        fmax = 100
        kmax = 50
        theta = 0.5
        tat = 1
        d = 1
        fini = 28

        mode_obj = TransportMode(name, bya, co, c1, c2, v, t, fmax, kmax, theta, tat, d, fini)

        m.add_mode(mode_obj)

        self.assertEqual(len(m.get_modes()), 3)

        with self.assertRaises(exceptions.AddModeExceptions):
            mode_obj = TransportMode("bus", bya, co, c1, c2, v, t, fmax, kmax, theta, tat, d, fini)
            m.add_mode(mode_obj)

        with self.assertRaises(exceptions.ModeIsNotValidException):
            m.add_mode("mode_obj")

    def test_delete_mode(self):
        """
        to test delete_mode method
        :return:
        """

        m = TransportModeManager()

        m.remove_mode("bus")
        self.assertEqual(len(m.get_modes()), 1)

        with self.assertRaises(exceptions.ModeDoesNotExistExceptions):
            m.remove_mode("train")


if __name__ == '__main__':
    unittest.main()



from collections import defaultdict
from typing import List

from sidermit.publictransportsystem import Route

defaultdict_float = defaultdict(float)
defaultdict2_float = defaultdict(lambda: defaultdict(float))
defaultdict3_float = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))


class OperatorsCost:

    @staticmethod
    def lines_travel_time(routes: List[Route], edge_distance: defaultdict2_float) -> defaultdict_float:
        """
        to get a dictionary with travel times for all lines defined in the network
        :param edge_distance: dict with edge distance [m], with 2 keys dic[nodei_id][nodej_id]= distance [m]
        :param routes: list of Route object
        :return: dict with time on board [hr] of vehicle for each route_id
        """
        line_travel_time = defaultdict(float)

        for route in routes:
            node_sequence_i = route.nodes_sequence_i
            node_sequence_r = route.nodes_sequence_r

            tv = 0

            # recorremos secuencia de ida
            for i in range(len(node_sequence_i) - 1):
                j = i + 1

                nodei_id = node_sequence_i[i]
                nodej_id = node_sequence_i[j]

                tv += edge_distance[nodei_id][nodej_id] / route.mode.v

            # recorremos secuencia de vuelta
            for i in range(len(node_sequence_r) - 1):
                j = i + 1

                nodei_id = node_sequence_r[i]
                nodej_id = node_sequence_r[j]

                tv += edge_distance[nodei_id][nodej_id] / route.mode.v

            line_travel_time[route.id] = tv

        return line_travel_time

    @staticmethod
    def get_cycle_time(z: defaultdict3_float, v: defaultdict3_float, routes: List[Route],
                       line_travel_time: defaultdict_float) -> defaultdict_float:
        """
        to get cycle time of all routes in the network
        :param z: boarding, dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :param v: alighting, dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :param routes: list of Route object
        :param line_travel_time: dict with time on board of vehicle for each route_id
        :return: dict with cycle time [hr] for each route_id
        """

        cycle_time = defaultdict(float)

        for route in routes:
            route_id = route.id
            t = route.mode.t / 3600

            tc = line_travel_time[route_id]

            # secuancial
            if route.mode.bya == 0:

                for stop in z[route_id]["I"]:
                    tc += t * z[route_id]["I"][stop]
                for stop in v[route_id]["I"]:
                    tc += t * v[route_id]["I"][stop]
                for stop in z[route_id]["R"]:
                    tc += t * z[route_id]["R"][stop]
                for stop in v[route_id]["R"]:
                    tc += t * v[route_id]["R"][stop]

            # simultaneo
            if route.mode.bya == 1:
                processed_stop_i = []
                processed_stop_r = []

                for stop in z[route_id]["I"]:
                    processed_stop_i.append(stop)
                    tc += t * max(z[route_id]["I"][stop], v[route_id]["I"][stop])
                for stop in v[route_id]["I"]:
                    if stop not in processed_stop_i:
                        tc += t * max(z[route_id]["I"][stop], v[route_id]["I"][stop])
                for stop in z[route_id]["R"]:
                    processed_stop_r.append(stop)
                    tc += t * max(z[route_id]["R"][stop], v[route_id]["R"][stop])
                for stop in v[route_id]["R"]:
                    if stop not in processed_stop_r:
                        tc += t * max(z[route_id]["R"][stop], v[route_id]["R"][stop])

            cycle_time[route_id] = tc

        return cycle_time

    @staticmethod
    def get_operators_cost(routes: List[Route], cycle_time: defaultdict_float, f: defaultdict_float,
                           k: defaultdict_float) -> float:
        """
        to get operators cost give a frequencies and boarding size for all lines
        :param routes: list Route object
        :param cycle_time: dict with cycle time [hr] for each route_id
        :param f: dic[route_id] = frequency [veh/hr]
        :param k: dic[route_id] = frequency [pax/veh]
        :return: float, operators cost
        """

        CO = 0

        for route in routes:
            route_id = route.id

            f_r = f[route_id]

            if f_r == 0:
                continue

            k_r = k[route_id]
            tc_r = cycle_time[route_id]

            mode = route.mode
            c0 = mode.co
            c1 = mode.c1

            CO += (c0 + c1 * k_r) * f_r * tc_r

        return CO



from setuptools import setup, find_packages

setup(
    name='sidermit',
    packages=find_packages(exclude=['tests', 'examples']),
    version='0.0.10',
    license='gpl-3.0',
    description='optimize public transport system on a city graph based on one CBD and zones',
    author='Felipe Vera',
    author_email='fvera@transapp.cl',
    url='https://github.com/SIDERMIT/optimizer',
    download_url='https://github.com/SIDERMIT/optimizer/archive/v0.0.10.tar.gz',
    keywords=['public transport', 'optimize'],
    install_requires=[
        'pandas>=1.0.5',
        'matplotlib>=3.3.0',
        'networkx>=2.4',
        'scipy>=1.5.2',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7'
    ]
)



from .demand import Demand
from .graph import CBD, Periphery, Subcenter, Graph, GraphContentFormat, Node

__all__ = ['CBD', 'Periphery', 'Subcenter', 'Graph', 'GraphContentFormat',
           'Demand', 'Node']



from collections import defaultdict

from sidermit.city import Graph
from sidermit.publictransportsystem import TransportNetwork

defaultdict_float = defaultdict(float)


class InfrastructureCost:
    @staticmethod
    def get_mode_network_distance(graph_obj: Graph, network_obj: TransportNetwork,
                                  f: defaultdict_float) -> defaultdict_float:
        """
        to get total distance builded in each transport mode
        :param network_obj: TransportNetwork object
        :param graph_obj: Graph object
        :param f: dict with frequency for each route_id
        :return: ddict with total distance for each mode in transport network
        """
        edges = graph_obj.get_edges()
        edges_distance = graph_obj.get_edges_distance()

        routes = network_obj.get_routes()
        list_modes = network_obj.get_modes()

        mode_distance = defaultdict(float)

        edge_list = []

        for edge in edges:
            d_e = edges_distance[edge.node1.id][edge.node2.id]
            for mode in list_modes:
                for route in routes:
                    if route.mode == mode:
                        if f[route.id] != 0:
                            node_sequence_i = route.nodes_sequence_i
                            node_sequence_r = route.nodes_sequence_r

                            ver_sum = False
                            for i in range(len(node_sequence_i) - 1):
                                j = i + 1
                                if node_sequence_i[i] == edge.node1.id and node_sequence_i[j] == edge.node2.id and (
                                edge.node1.id, edge.node2.id) not in edge_list:
                                    mode_distance[mode] += d_e * mode.d
                                    edge_list.append((edge.node1.id, edge.node2.id))
                                    edge_list.append((edge.node2.id, edge.node1.id))
                                    ver_sum = True
                                    break
                            if ver_sum is True:
                                break
                            for i in range(len(node_sequence_r) - 1):
                                j = i + 1
                                if node_sequence_r[i] == edge.node1.id and node_sequence_r[j] == edge.node2.id and (
                                edge.node1.id, edge.node2.id) not in edge_list:
                                    mode_distance[mode] += d_e * mode.d
                                    edge_list.append((edge.node1.id, edge.node2.id))
                                    edge_list.append((edge.node2.id, edge.node1.id))
                                    ver_sum = True
                                    break
                            if ver_sum is True:
                                break

        return mode_distance

    @staticmethod
    def get_infrastruture_cost(graph_obj: Graph, network_obj: TransportNetwork, f: defaultdict_float) -> float:
        """
        to get infrastruture cost
        :param network_obj: TransportNetwork object
        :param graph_obj: Graph object
        :param f: dict with frequency for each route_id
        :return: infrastruture cost
        """

        infrastruture_cost_obj = InfrastructureCost()

        mode_distance = infrastruture_cost_obj.get_mode_network_distance(graph_obj, network_obj, f)

        CI = 0

        for mode in mode_distance:
            CI += mode.c2 * mode_distance[mode]
        return CI



from __future__ import annotations

from collections import defaultdict

import pandas as pd

from sidermit.city.graph import Graph, CBD, Periphery, Subcenter
from sidermit.exceptions import *

defaultdict2_float = defaultdict(lambda: defaultdict(float))


class Demand:

    def __init__(self, graph_obj: Graph):
        """
        class that allows to build the OD matrix
        :param graph_obj: city graph object, necessary to recognize the id of the created nodes and compatibility
        with that of the OD matrix
        """
        # initialization of city graph
        self.__graph_obj = graph_obj
        # number of zones
        self.__n = self.__graph_obj.get_n()
        # matrix with demand information
        # dictionary structure [node_idO][node_idD] = vij
        self.__matrix = defaultdict(lambda: defaultdict(float))
        # total trips
        self.total_trips = 0
        # initialization of matriz with zero trips in all OD pairs
        self.__build_default_matrix()

    def get_total_trips(self) -> float:
        """
        to get total trips in all OD pair
        :return:
        """
        return self.total_trips

    def matrix_to_file(self, file_path):
        """
        write OD matrix defined in a csv file
        :param file_path:
        :return:
        """
        col_origin = []
        col_destination = []
        col_vij = []

        for origin in self.__matrix:
            for destination in self.__matrix[origin]:
                col_origin.append(origin)
                col_destination.append(destination)
                col_vij.append(self.__matrix[origin][destination])

        df_matrix = pd.DataFrame()
        df_matrix["origin"] = col_origin
        df_matrix["destination"] = col_destination
        df_matrix["vij"] = col_vij

        df_matrix.to_csv(file_path, sep=",", index=False)

    def __build_default_matrix(self):
        """
        Build initial OD matrix with 0 trips in all OD pairs
        :return: None
        """
        for origin_node in self.__graph_obj.get_nodes():
            for destination_node in self.__graph_obj.get_nodes():
                self.__matrix[int(origin_node.id)][int(destination_node.id)] = 0

    def change_vij(self, origin_node_id: int, destination_node_id: int, vij: float):
        """
        Change trip value to a OD pair
        :param origin_node_id:
        :param destination_node_id:
        :param vij: trips in OD pair
        :return:
        """

        if vij < 0:
            raise TripsValueIsNotValidException("trips value must be >= 0")

        if self.__matrix.get(origin_node_id):
            if destination_node_id in self.__matrix[origin_node_id]:
                self.total_trips = self.total_trips - self.__matrix[origin_node_id][destination_node_id] + vij
                self.__matrix[origin_node_id][destination_node_id] = vij
            else:
                raise DestinationIdDoesNotFoundException("id destination does not found")
        else:
            raise OriginIdDoesNotFoundException("id origin does not found")

    def get_matrix(self) -> defaultdict2_float:
        """
        Get last OD matrix saved
        :return:
        """
        return self.__matrix

    @staticmethod
    def parameters_validator(y: float, a: float, alpha: float, beta: float) -> bool:
        """
        to validate construction parameters of symmetric OD matrix
        :param y: total demand
        :param a: proportion of trips generated in the peripheries
        :param alpha: proportion of trips from the periphery to the CBD
        :param beta: proportion of trips from the periphery that go to the sub-center of the same zone
        :return: True if parameters are valid. False if not.
        """
        if not isinstance(y, int) and not isinstance(y, float):
            raise YIsNotValidException("must specify value int or float for y")
        if not isinstance(a, int) and not isinstance(a, float):
            raise AIsNotValidException("must specify value int or float for a")
        if not isinstance(alpha, int) and not isinstance(alpha, float):
            raise AlphaIsNotValidException("must specify value int or float for alpha")
        if not isinstance(beta, int) and not isinstance(beta, float):
            raise BetaIsNotValidException("must specify value int or float for beta")
        if y < 0:
            raise YOutOfRangeException("y must be positive")
        if a > 1 or a < 0:
            raise AOutOfRangeException("a must be in range [0-1]")
        if alpha > 1 or alpha < 0:
            raise AlphaOutOfRangeException("alpha must be in range [0-1]")
        if beta >= 1 or beta < 0:
            raise BetaOutOfRangeException("beta must be in range [0-1)")
        if alpha + beta > 1:
            raise AlphaBetaOutOfRangeException("alpha + beta must be in range [0-1]")

        return True

    @staticmethod
    def build_from_parameters(graph_obj: Graph, y: float, a: float, alpha: float, beta: float) -> Demand:
        """
        to build OD matrix with symmetric parameters
        :param graph_obj: city graph object, necessary to recognize the id of the created nodes and compatibility
        with that of the OD matrix.
        :param y: total demand
        :param a: proportion of trips generated in the peripheries
        :param alpha: proportion of trips from the periphery to the CBD
        :param beta: proportion of trips from the periphery that go to the sub-center of the same zone
        :return:
        """
        demand_obj = Demand(graph_obj)

        if demand_obj.parameters_validator(y, a, alpha, beta):

            b = 1 - a
            gamma = 0
            gamma_g = 0
            n = demand_obj.__n

            if n > 1:
                gamma = 1 - alpha - beta
                alpha_g = alpha / (1 - beta)
                gamma_g = gamma / (1 - beta)
            else:
                beta = 1 - alpha
                alpha_g = 1

            y_z = y
            # demand for each zone
            if n != 0:
                y_z = y / n
            # periphery demand
            y_p = a * y_z
            # subcenter demand
            y_sc = b * y_z
            # trips periphery - CBD
            v_p_cbd = alpha * y_p
            # trips periphery - subcenter
            v_p_sc = beta * y_p
            # trips periphery - other subcenter
            v_p_osc = 0
            if n > 1:
                v_p_osc = gamma * y_p / (n - 1)
            # trips subcenter - cbd
            v_sc_cbd = y_sc * alpha_g
            # trips subcenter -  other subcenter
            v_sc_osc = 0
            if n > 1:
                v_sc_osc = y_sc * gamma_g / (n - 1)

            # for each origin in matrix
            for origin_node_id in demand_obj.__matrix:
                # for each destination in matrix[origin_id]
                for destination_node_id in demand_obj.__matrix[origin_node_id]:
                    origin_node_obj = None
                    destination_node_obj = None
                    for node in demand_obj.__graph_obj.get_nodes():
                        if node.id == origin_node_id:
                            origin_node_obj = node
                        if node.id == destination_node_id:
                            destination_node_obj = node

                    # CBD does not generate trips
                    if isinstance(origin_node_obj, CBD):
                        continue
                    # periphery does not attract
                    if isinstance(destination_node_obj, Periphery):
                        continue
                    # trips between P - CBD
                    if isinstance(origin_node_obj, Periphery) and isinstance(destination_node_obj, CBD):
                        demand_obj.change_vij(origin_node_id, destination_node_id, v_p_cbd)
                    # trips between P - SC
                    if isinstance(origin_node_obj, Periphery) and isinstance(destination_node_obj, Subcenter) and \
                            origin_node_obj.zone_id == destination_node_obj.zone_id:
                        demand_obj.change_vij(origin_node_id, destination_node_id, v_p_sc)
                    # trips between P - OSC
                    if isinstance(origin_node_obj, Periphery) and isinstance(destination_node_obj, Subcenter) and \
                            origin_node_obj.zone_id != destination_node_obj.zone_id:
                        demand_obj.change_vij(origin_node_id, destination_node_id, v_p_osc)
                    # trips between SC - CBD
                    if isinstance(origin_node_obj, Subcenter) and isinstance(destination_node_obj, CBD):
                        demand_obj.change_vij(origin_node_id, destination_node_id, v_sc_cbd)
                    # trips between SC - OSC
                    if isinstance(origin_node_obj, Subcenter) and isinstance(destination_node_obj, Subcenter) and \
                            origin_node_obj.zone_id != destination_node_obj.zone_id:
                        demand_obj.change_vij(origin_node_id, destination_node_id, v_sc_osc)

        return demand_obj

    @staticmethod
    def build_from_file(graph_obj: Graph, file_path) -> Demand:
        """
        to build OD matrix from file
        :param graph_obj: city graph object, necessary to recognize the id of the created nodes and compatibility
        with that of the OD matrix.
        :param file_path:
        :return:
        """

        demand_obj = Demand(graph_obj)

        with open(file_path, mode='r', encoding='utf-8') as f_obj:

            n_lines = 0
            for line in f_obj.readlines():
                if n_lines == 0:
                    n_lines = 1
                    continue
                else:
                    if len(line.split(",")) == 3:
                        origin_id, destination_id, vij = line.split(",")
                        try:
                            demand_obj.change_vij(int(origin_id), int(destination_id), float(vij))
                        except ValueError:
                            raise NodeIdDemandIsNotAnInteger("Node id must be a integer")
                    else:
                        raise FileFormatIsNotValidException("each line must provide information about [origin_id] ["
                                                            "destination_id] [vij]")
        return demand_obj

    @staticmethod
    def build_from_content(graph_obj: Graph, matrix) -> Demand:
        """

        :param graph_obj: city graph object, necessary to recognize the id of the created nodes and compatibility
        with that of the OD matrix.
        :param matrix: List[List[float]]
        :return: Demand object
        """

        demand_obj = Demand(graph_obj)
        nodes = graph_obj.get_nodes()

        if len(nodes) != len(matrix):
            raise DemandMatrixIsNotValidException("Matrix should have rows equal to number of nodes")
        for i in range(len(matrix)):
            if len(nodes) != len(matrix[i]):
                raise DemandMatrixIsNotValidException("Matrix should have columns equal to number of nodes")
            for j in range(len(matrix[i])):
                origin_node_id = nodes[i].id
                destination_node_id = nodes[j].id
                vij = matrix[i][j]
                if vij < 0:
                    raise VijIsNotValidException("Vij should be >=0")

                demand_obj.change_vij(origin_node_id, destination_node_id, vij)

        return demand_obj
